import enum

import zigpy.config
import zigpy.types as t


class Repr:
    def __repr__(self) -> str:
        r = "<%s " % (self.__class__.__name__,)
        r += " ".join(["%s=%s" % (f, getattr(self, f, None)) for f in vars(self)])
        r += ">"
        return r


class Timeouts:
    SREQ = 6000
    reset = 30000
    default = 10000


class AddressMode(t.uint8_t, enum.Enum):
    ADDR_NOT_PRESENT = 0
    ADDR_GROUP = 1
    ADDR_16BIT = 2
    ADDR_64BIT = 3
    ADDR_BROADCAST = 15


class LedMode(t.uint8_t, enum.Enum):
    Off = 0
    On = 1
    Blink = 2
    Flash = 3
    Toggle = 4


class ZnpVersion(t.uint8_t, enum.Enum):
    zStack12 = 0
    zStack3x0 = 1
    zStack30x = 2


class CommandType(t.uint8_t, enum.Enum):
    POLL = 0
    SREQ = 1
    AREQ = 2
    SRSP = 3


class Subsystem(t.uint8_t, enum.Enum):
    RESERVED = 0
    SYS = 1
    MAC = 2
    NWK = 3
    AF = 4
    ZDO = 5
    SAPI = 6
    UTIL = 7
    DEBUG = 8
    APP = 9
    APP_CNF = 15
    GREENPOWER = 21


class ParameterType(t.uint8_t, enum.Enum):
    UINT8 = 0
    UINT16 = 1
    UINT32 = 2
    IEEEADDR = 3

    BUFFER = 4
    BUFFER8 = 5
    BUFFER16 = 6
    BUFFER18 = 7
    BUFFER32 = 8
    BUFFER42 = 9
    BUFFER100 = 10

    LIST_UINT8 = 11
    LIST_UINT16 = 12
    LIST_ROUTING_TABLE = 13
    LIST_BIND_TABLE = 14
    LIST_NEIGHBOR_LQI = 15
    LIST_NETWORK = 16
    LIST_ASSOC_DEV = 17

    INT8 = 18

    def is_buffer(type):
        return (
            type == ParameterType.BUFFER
            or type == ParameterType.BUFFER8
            or type == ParameterType.BUFFER16
            or type == ParameterType.BUFFER18
            or type == ParameterType.BUFFER32
            or type == ParameterType.BUFFER42
            or type == ParameterType.BUFFER100
        )


class NetworkOptions(Repr):
    networkKey: t.KeyData
    panID: t.PanId
    extendedPanID: t.ExtendedPanId
    channelList: t.Channels

    def __init__(self, config: zigpy.config.SCHEMA_NETWORK) -> None:
        self.networkKeyDistribute = False
        self.networkKey = config[zigpy.config.CONF_NWK_KEY] or (
            zigpy.config.cv_key([1, 3, 5, 7, 9, 11, 13, 15, 0, 2, 4, 6, 8, 10, 12, 13])
        )
        self.panID = config[zigpy.config.CONF_NWK_PAN_ID] or t.PanId(0x1A62)
        self.extendedPanID = config[zigpy.config.CONF_NWK_EXTENDED_PAN_ID] or (
            t.ExtendedPanId([0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD, 0xDD])
        )
        self.channelList = (
            config[zigpy.config.CONF_NWK_CHANNELS] or t.Channels.CHANNEL_11
        )



# This file updated automatically before PyPI release was created
MAJOR_VERSION = 0
MINOR_VERSION = 5
PATCH_VERSION = "0-beta"
__short_version__ = "{}.{}".format(MAJOR_VERSION, MINOR_VERSION)
__version__ = "{}.{}".format(__short_version__, PATCH_VERSION)



import voluptuous as vol
from zigpy.config import (  # noqa: F401 pylint: disable=unused-import
    CONF_DATABASE,
    CONF_DEVICE,
    CONF_DEVICE_PATH,
    CONFIG_SCHEMA,
    cv_boolean,
)

CONF_DEVICE_BAUDRATE = "baudrate"
CONF_DEVICE_BAUDRATE_DEFAULT = 115200
CONF_FLOW_CONTROL = "flow_control"
CONF_FLOW_CONTROL_DEFAULT = None

SCHEMA_DEVICE = vol.Schema(
    {
        vol.Required(CONF_DEVICE_PATH): vol.Any(vol.PathExists(), "auto"),
        vol.Optional(CONF_DEVICE_BAUDRATE, default=CONF_DEVICE_BAUDRATE_DEFAULT): int,
        vol.Optional(CONF_FLOW_CONTROL, default=CONF_FLOW_CONTROL_DEFAULT): vol.In(
            ("hardware", "software", None)
        ),
    }
)

CONFIG_SCHEMA = CONFIG_SCHEMA.extend({vol.Required(CONF_DEVICE): SCHEMA_DEVICE})



import logging

from zigpy.exceptions import ZigbeeException

LOGGER = logging.getLogger(__name__)


class TODO(Exception):
    def __init__(self, msg, *args, **kwargs):
        super().__init__(msg, *args, **kwargs)
        LOGGER.warning("Not implemented: " + msg, *args)


class CommandError(ZigbeeException):
    def __init__(self, status, *args, **kwargs):
        self._status = status
        super().__init__(*args, **kwargs)

    @property
    def status(self):
        return self._status



import enum


class ResetType(enum.IntEnum):
    HARD = 0
    SOFT = 1


class System:
    resetType = ResetType


class DeviceLogicalType(enum.IntEnum):
    COORDINATOR = 0
    ROUTER = 1
    ENDDEVICE = 2
    COMPLEX_DESC_AVAIL = 4
    USER_DESC_AVAIL = 8
    RESERVED1 = 16
    RESERVED2 = 32
    RESERVED3 = 64
    RESERVED4 = 128


class ZDO:
    deviceLogicalType = DeviceLogicalType


class networkLatencyReq(enum.IntEnum):
    NO_LATENCY_REQS = 0
    FAST_BEACONS = 1
    SLOW_BEACONS = 2


class AF:
    networkLatencyReq = networkLatencyReq


class Constants:
    AF = AF
    SYS = System
    ZDO = ZDO



import asyncio
import logging
from asyncio.locks import Semaphore
from typing import Any, Dict, Optional

import zigpy.application
import zigpy.config
import zigpy.device
import zigpy.types
import zigpy.util
from zigpy.profiles import zha
from zigpy.types import BroadcastAddress
from zigpy.zdo.types import ZDOCmd

from zigpy_cc import __version__, types as t
from zigpy_cc.api import API
from zigpy_cc.config import CONF_DEVICE, CONFIG_SCHEMA, SCHEMA_DEVICE
from zigpy_cc.exception import TODO, CommandError
from zigpy_cc.types import NetworkOptions, Subsystem, ZnpVersion, LedMode, AddressMode
from zigpy_cc.zigbee.start_znp import start_znp
from zigpy_cc.zpi_object import ZpiObject

LOGGER = logging.getLogger(__name__)

CHANGE_NETWORK_WAIT = 1
SEND_CONFIRM_TIMEOUT = 60
PROTO_VER_WATCHDOG = 0x0108

REQUESTS = {
    "nwkAddrReq": (ZDOCmd.NWK_addr_req, 0),
    "ieeeAddrReq": (ZDOCmd.IEEE_addr_req, 0),
    "matchDescReq": (ZDOCmd.Match_Desc_req, 2),
    "endDeviceAnnceInd": (ZDOCmd.Device_annce, 2),
    "mgmtLqiRsp": (ZDOCmd.Mgmt_Lqi_rsp, 2),
    "mgmtPermitJoinReq": (ZDOCmd.Mgmt_Permit_Joining_req, 3),
    "mgmtPermitJoinRsp": (ZDOCmd.Mgmt_Permit_Joining_rsp, 2),
    "nodeDescRsp": (ZDOCmd.Node_Desc_rsp, 2),
    "activeEpRsp": (ZDOCmd.Active_EP_rsp, 2),
    "simpleDescRsp": (ZDOCmd.Simple_Desc_rsp, 2),
    "bindRsp": (ZDOCmd.Bind_rsp, 2),
}

IGNORED = (
    "bdbComissioningNotifcation",
    "dataConfirm",
    "leaveInd",
    "srcRtgInd",
    "stateChangeInd",
    "tcDeviceInd",
)


class ControllerApplication(zigpy.application.ControllerApplication):
    _semaphore: Semaphore
    _api: Optional[API]
    SCHEMA = CONFIG_SCHEMA
    SCHEMA_DEVICE = SCHEMA_DEVICE

    probe = API.probe

    def __init__(self, config: Dict[str, Any]):
        super().__init__(config=zigpy.config.ZIGPY_SCHEMA(config))
        self._api = None

        self.discovering = False
        self.version = {}

    async def shutdown(self):
        """Shutdown application."""
        self._api.close()

    def connection_lost(self):
        asyncio.create_task(self.reconnect())

    async def reconnect(self):
        while True:
            if self._api is not None:
                self._api.close()
            await asyncio.sleep(5)
            try:
                await self.startup(True)
                break
            except Exception as e:
                LOGGER.info("Failed to reconnect: %s", e)
                pass

    async def startup(self, auto_form=False):
        """Perform a complete application startup"""

        LOGGER.info("Starting zigpy-cc version: %s", __version__)
        self._api = await API.new(self, self._config[CONF_DEVICE])

        try:
            await self._api.request(Subsystem.SYS, "ping", {"capabilities": 1})
        except CommandError as e:
            raise Exception("Failed to connect to the adapter(%s)", e)

        self.version = await self._api.version()

        concurrent = 16 if self.version["product"] == ZnpVersion.zStack3x0 else 2
        LOGGER.debug("Adapter concurrent: %d", concurrent)

        self._semaphore = asyncio.Semaphore(concurrent)

        ver = ZnpVersion(self.version["product"]).name
        LOGGER.info("Detected znp version '%s' (%s)", ver, self.version)

        if auto_form:
            await self.form_network()

        data = await self._api.request(Subsystem.UTIL, "getDeviceInfo", {})
        self._ieee = data.payload["ieeeaddr"]
        self._nwk = data.payload["shortaddr"]

        # add coordinator
        self.devices[self._ieee] = Coordinator(self, self._ieee, self._nwk)

    async def permit_with_key(self, node, code, time_s=60):
        raise TODO("permit_with_key")

    async def force_remove(self, dev: zigpy.device.Device):
        """Forcibly remove device from NCP."""
        LOGGER.warning("FORCE REMOVE %s", dev)

    async def form_network(self, channel=15, pan_id=None, extended_pan_id=None):
        LOGGER.info("Forming network")
        LOGGER.debug("Config: %s", self.config)
        options = NetworkOptions(self.config[zigpy.config.CONF_NWK])
        LOGGER.debug("NetworkOptions: %s", options)
        backupPath = ""
        status = await start_znp(
            self._api, self.version["product"], options, 0x0B84, backupPath
        )
        LOGGER.info("ZNP started, status: %s", status)

        self.set_led(LedMode.Off)

    async def mrequest(
        self,
        group_id,
        profile,
        cluster,
        src_ep,
        sequence,
        data,
        *,
        hops=0,
        non_member_radius=3
    ):
        """Submit and send data out as a multicast transmission.

        :param group_id: destination multicast address
        :param profile: Zigbee Profile ID to use for outgoing message
        :param cluster: cluster id where the message is being sent
        :param src_ep: source endpoint id
        :param sequence: transaction sequence number of the message
        :param data: Zigbee message payload
        :param hops: the message will be delivered to all nodes within this number of
                     hops of the sender. A value of zero is converted to MAX_HOPS
        :param non_member_radius: the number of hops that the message will be forwarded
                                  by devices that are not members of the group. A value
                                  of 7 or greater is treated as infinite
        :returns: return a tuple of a status and an error_message. Original requestor
                  has more context to provide a more meaningful error message
        """
        LOGGER.debug(
            "multicast %s",
            (
                group_id,
                profile,
                cluster,
                src_ep,
                sequence,
                data,
                hops,
                non_member_radius,
            ),
        )
        try:
            obj = ZpiObject.from_cluster(
                group_id,
                profile,
                cluster,
                src_ep or 1,
                0xFF,
                sequence,
                data,
                addr_mode=AddressMode.ADDR_GROUP,
            )
            waiter_id = None
            waiter = self._api.create_response_waiter(obj, sequence)
            if waiter:
                waiter_id = waiter.id

            async with self._semaphore:
                await self._api.request_raw(obj, waiter_id)
                """
                As a group command is not confirmed and thus immediately returns
                (contrary to network address requests) we will give the
                command some time to 'settle' in the network.
                """
                await asyncio.sleep(0.2)

        except CommandError as ex:
            return ex.status, "Couldn't enqueue send data multicast: {}".format(ex)

        return 0, "message send success"

    @zigpy.util.retryable_request
    async def request(
        self,
        device,
        profile,
        cluster,
        src_ep,
        dst_ep,
        sequence,
        data,
        expect_reply=True,
        use_ieee=False,
    ):
        LOGGER.debug(
            "request %s",
            (
                device.nwk,
                profile,
                cluster,
                src_ep,
                dst_ep,
                sequence,
                data,
                expect_reply,
                use_ieee,
            ),
        )

        try:
            obj = ZpiObject.from_cluster(
                device.nwk, profile, cluster, src_ep, dst_ep, sequence, data
            )
            waiter_id = None
            if expect_reply:
                waiter = self._api.create_response_waiter(obj, sequence)
                if waiter:
                    waiter_id = waiter.id

            async with self._semaphore:
                await self._api.request_raw(obj, waiter_id)

        except CommandError as ex:
            return ex.status, "Couldn't enqueue send data request: {}".format(ex)

        return 0, "message send success"

    async def broadcast(
        self,
        profile,
        cluster,
        src_ep,
        dst_ep,
        grpid,
        radius,
        sequence,
        data,
        broadcast_address=zigpy.types.BroadcastAddress.RX_ON_WHEN_IDLE,
    ):
        LOGGER.debug(
            "broadcast %s",
            (
                profile,
                cluster,
                src_ep,
                dst_ep,
                grpid,
                radius,
                sequence,
                data,
                broadcast_address,
            ),
        )
        try:
            obj = ZpiObject.from_cluster(
                broadcast_address,
                profile,
                cluster,
                src_ep,
                dst_ep,
                sequence,
                data,
                radius=radius,
                addr_mode=AddressMode.ADDR_16BIT,
            )

            async with self._semaphore:
                await self._api.request_raw(obj)
                """
                As a broadcast command is not confirmed and thus immediately returns
                (contrary to network address requests) we will give the
                command some time to 'settle' in the network.
                """
                await asyncio.sleep(0.2)

        except CommandError as ex:
            return (
                ex.status,
                "Couldn't enqueue send data request for broadcast: {}".format(ex),
            )

        return 0, "broadcast send success"

    async def permit_ncp(self, time_s=60):
        assert 0 <= time_s <= 254
        payload = {
            "addrmode": AddressMode.ADDR_BROADCAST,
            "dstaddr": BroadcastAddress.ALL_ROUTERS_AND_COORDINATOR,
            "duration": time_s,
            "tcsignificance": 0,
        }
        async with self._semaphore:
            await self._api.request(Subsystem.ZDO, "mgmtPermitJoinReq", payload)

    def handle_znp(self, obj: ZpiObject):
        if obj.command_type != t.CommandType.AREQ:
            return

        frame = obj.to_unpi_frame()

        nwk = obj.payload["srcaddr"] if "srcaddr" in obj.payload else None
        ieee = obj.payload["ieeeaddr"] if "ieeeaddr" in obj.payload else None
        profile_id = zha.PROFILE_ID
        src_ep = 0
        dst_ep = 0
        lqi = 0
        rssi = 0

        if obj.subsystem == t.Subsystem.ZDO and obj.command == "endDeviceAnnceInd":
            nwk = obj.payload["nwkaddr"]
            LOGGER.info("New device joined: 0x%04x, %s", nwk, ieee)
            self.handle_join(nwk, ieee, 0)
            obj.sequence = 0

        if obj.command in IGNORED:
            return

        if obj.subsystem == t.Subsystem.ZDO and obj.command == "mgmtPermitJoinRsp":
            self.set_led(LedMode.On)

        if obj.subsystem == t.Subsystem.ZDO and obj.command == "permitJoinInd":
            self.set_led(LedMode.Off if obj.payload["duration"] == 0 else LedMode.On)
            return

        if obj.subsystem == t.Subsystem.ZDO and obj.command in REQUESTS:
            if obj.sequence is None:
                return
            LOGGER.debug("REPLY for %d %s", obj.sequence, obj.command)
            cluster_id, prefix_length = REQUESTS[obj.command]
            tsn = bytes([obj.sequence])
            data = tsn + frame.data[prefix_length:]

        elif obj.subsystem == t.Subsystem.AF and (
            obj.command == "incomingMsg" or obj.command == "incomingMsgExt"
        ):
            # ZCL commands
            cluster_id = obj.payload["clusterid"]
            src_ep = obj.payload["srcendpoint"]
            dst_ep = obj.payload["dstendpoint"]
            data = obj.payload["data"]
            lqi = obj.payload["linkquality"]

        else:
            LOGGER.warning(
                "Unhandled message: %s %s %s",
                t.CommandType(obj.command_type),
                t.Subsystem(obj.subsystem),
                obj.command,
            )
            return

        try:
            if ieee:
                device = self.get_device(ieee=ieee)
            else:
                device = self.get_device(nwk=nwk)
        except KeyError:
            LOGGER.warning(
                "Received frame from unknown device: %s", ieee if ieee else nwk
            )
            return

        device.radio_details(lqi, rssi)

        LOGGER.info("handle_message %s", obj.command)
        self.handle_message(device, profile_id, cluster_id, src_ep, dst_ep, data)

    def set_led(self, mode: LedMode):
        if self.version["product"] != ZnpVersion.zStack3x0:
            try:
                loop = asyncio.get_event_loop()
                loop.create_task(
                    self._api.request(
                        Subsystem.UTIL, "ledControl", {"ledid": 3, "mode": mode}
                    )
                )
            except Exception as ex:
                LOGGER.warning("Can't set LED: %s", ex)


class Coordinator(zigpy.device.Device):
    """
    todo add endpoints?
    @see zStackAdapter.ts - getCoordinator
    """

    @property
    def manufacturer(self):
        return "Texas Instruments"

    @property
    def model(self):
        return "ZNP Coordinator"



from zigpy.types import ExtendedPanId, Channels

from zigpy_cc.types import ZnpVersion
from zigpy_cc.zigbee.common import NvItemsIds


class Items:
    @staticmethod
    def znpHasConfiguredInit(version):
        return {
            "id": NvItemsIds.ZNP_HAS_CONFIGURED_ZSTACK1
            if version == ZnpVersion.zStack12
            else NvItemsIds.ZNP_HAS_CONFIGURED_ZSTACK3,
            "len": 0x01,
            "initlen": 0x01,
            "initvalue": b"\x00",
        }

    @staticmethod
    def znpHasConfigured(version):
        return {
            "id": NvItemsIds.ZNP_HAS_CONFIGURED_ZSTACK1
            if version == ZnpVersion.zStack12
            else NvItemsIds.ZNP_HAS_CONFIGURED_ZSTACK3,
            "offset": 0x00,
            "len": 0x01,
            "value": b"\x55",
        }

    @staticmethod
    def panID(panID):
        return {
            "id": NvItemsIds.PANID,
            "len": 0x02,
            "offset": 0x00,
            "value": bytes([panID & 0xFF, (panID >> 8) & 0xFF]),
        }

    @staticmethod
    def extendedPanID(extendedPanID: ExtendedPanId):
        return {
            "id": NvItemsIds.EXTENDED_PAN_ID,
            "len": 0x08,
            "offset": 0x00,
            "value": extendedPanID.serialize(),
        }

    @staticmethod
    def channelList(channelList: Channels):
        return {
            "id": NvItemsIds.CHANLIST,
            "len": 0x04,
            "offset": 0x00,
            "value": channelList.serialize(),
        }

    @staticmethod
    def networkKeyDistribute(distribute):
        return {
            "id": NvItemsIds.PRECFGKEYS_ENABLE,
            "len": 0x01,
            "offset": 0x00,
            "value": b"\x01" if distribute else b"\x00",
        }

    @staticmethod
    def networkKey(key):
        return {
            # id/configid is used depending if SAPI or SYS command is executed
            "id": NvItemsIds.PRECFGKEY,
            "configid": NvItemsIds.PRECFGKEY,
            "len": 0x10,
            "offset": 0x00,
            "value": bytes(key),
        }

    @staticmethod
    def startupOption(value):
        return {
            "id": NvItemsIds.STARTUP_OPTION,
            "len": 0x01,
            "offset": 0x00,
            "value": bytes([value]),
        }

    @staticmethod
    def logicalType(value):
        return {
            "id": NvItemsIds.LOGICAL_TYPE,
            "len": 0x01,
            "offset": 0x00,
            "value": bytes([value]),
        }

    @staticmethod
    def zdoDirectCb():
        return {
            "id": NvItemsIds.ZDO_DIRECT_CB,
            "len": 0x01,
            "offset": 0x00,
            "value": b"\x01",
        }

    @staticmethod
    def tcLinkKey():
        return {
            "id": NvItemsIds.LEGACY_TCLK_TABLE_START,
            "offset": 0x00,
            "len": 0x20,
            # ZigBee Alliance Pre-configured TC Link Key - 'ZigBeeAlliance09'
            "value": bytes(
                [
                    0xFF,
                    0xFF,
                    0xFF,
                    0xFF,
                    0xFF,
                    0xFF,
                    0xFF,
                    0xFF,
                    0x5A,
                    0x69,
                    0x67,
                    0x42,
                    0x65,
                    0x65,
                    0x41,
                    0x6C,
                    0x6C,
                    0x69,
                    0x61,
                    0x6E,
                    0x63,
                    0x65,
                    0x30,
                    0x39,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                    0x00,
                ]
            ),
        }



import asyncio
import logging
from typing import Any, Dict

import serial
import serial.tools.list_ports
import serial_asyncio
from serial.tools.list_ports_common import ListPortInfo

from zigpy_cc.config import CONF_DEVICE_BAUDRATE, CONF_DEVICE_PATH, CONF_FLOW_CONTROL
import zigpy_cc.types as t

LOGGER = logging.getLogger(__name__)

DataStart = 4
SOF = 0xFE

PositionDataLength = 1
PositionCmd0 = 2
PositionCmd1 = 3

MinMessageLength = 5
MaxDataSize = 250

"""
0451:     Texas Instruments
1a86:7523 QinHeng Electronics HL-340 USB-Serial adapter
          used in zzh - https://electrolama.com/projects/zig-a-zig-ah/
"""
usb_regexp = "0451:|1a86:7523"


class Parser:
    def __init__(self) -> None:
        self.buffer = b""

    def write(self, b: int):
        self.buffer += bytes([b])
        if SOF == self.buffer[0]:
            if len(self.buffer) > MinMessageLength:
                dataLength = self.buffer[PositionDataLength]

                fcsPosition = DataStart + dataLength
                frameLength = fcsPosition + 1

                if len(self.buffer) >= frameLength:
                    frameBuffer = self.buffer[0:frameLength]
                    self.buffer = self.buffer[frameLength:]

                    frame = UnpiFrame.from_buffer(dataLength, fcsPosition, frameBuffer)

                    return frame
        else:
            LOGGER.debug("drop char")
            self.buffer = b""

        return None


class UnpiFrame(t.Repr):
    def __init__(
        self,
        command_type: int,
        subsystem: int,
        command_id: int,
        data: bytes,
        length=None,
        fcs=None,
    ):
        self.command_type = t.CommandType(command_type)
        self.subsystem = t.Subsystem(subsystem)
        self.command_id = command_id
        self.data = data
        self.length = length
        self.fcs = fcs

    @classmethod
    def from_buffer(cls, length, fcs_position, buffer):
        subsystem = buffer[PositionCmd0] & 0x1F
        command_type = (buffer[PositionCmd0] & 0xE0) >> 5
        command_id = buffer[PositionCmd1]
        data = buffer[DataStart:fcs_position]
        fcs = buffer[fcs_position]

        checksum = cls.calculate_checksum(buffer[1:fcs_position])
        if checksum == fcs:
            return cls(command_type, subsystem, command_id, data, length, fcs)
        else:
            LOGGER.warning("Invalid checksum: 0x%s, data: 0x%s", checksum, buffer)
            return None

    @staticmethod
    def calculate_checksum(values):
        checksum = 0

        for value in values:
            checksum ^= value

        return checksum

    def to_buffer(self):
        length = len(self.data)
        cmd0 = ((self.command_type << 5) & 0xE0) | (self.subsystem & 0x1F)

        res = bytes([SOF, length, cmd0, self.command_id])
        res += self.data
        fcs = self.calculate_checksum(res[1:])

        return res + bytes([fcs])


class Gateway(asyncio.Protocol):
    _transport: serial_asyncio.SerialTransport

    def __init__(self, api, connected_future=None):
        self._parser = Parser()
        self._connected_future = connected_future
        self._api = api
        self._transport = None
        self._open = False

    def connection_made(self, transport: serial_asyncio.SerialTransport):
        """Callback when the uart is connected"""
        LOGGER.debug("Connection made")
        self._open = True
        self._transport = transport
        if self._connected_future:
            self._connected_future.set_result(True)

    def close(self):
        self._open = False
        self._transport.close()

    def write(self, data):
        self._transport.write(data)

    def send(self, frame: UnpiFrame):
        """Send data, taking care of escaping and framing"""
        data = frame.to_buffer()
        LOGGER.debug("Send: %s", data)
        self._transport.write(data)

    def data_received(self, data):
        """Callback when there is data received from the uart"""

        found = False
        for b in data:
            frame = self._parser.write(b)
            if frame is not None:
                found = True
                LOGGER.debug("Frame received: %s", frame)
                self._api.data_received(frame)

        if not found:
            LOGGER.info("Bytes received: %s", data)

    def connection_lost(self, exc):
        if self._open:
            LOGGER.error("Serial port closed unexpectedly: %s", exc)
            self._api.connection_lost()


def detect_port() -> ListPortInfo:
    devices = list(serial.tools.list_ports.grep(usb_regexp))
    if len(devices) < 1:
        raise serial.SerialException("Unable to find TI CC device using auto mode")
    if len(devices) > 1:
        raise serial.SerialException(
            "Unable to select TI CC device, multiple devices found: {}".format(
                ", ".join(map(lambda d: str(d), devices))
            )
        )
    return devices[0]


async def connect(config: Dict[str, Any], api, loop=None) -> Gateway:
    if loop is None:
        loop = asyncio.get_event_loop()

    connected_future = loop.create_future()
    protocol = Gateway(api, connected_future)

    port, baudrate = config[CONF_DEVICE_PATH], config[CONF_DEVICE_BAUDRATE]
    if port == "auto":
        device = detect_port()
        LOGGER.info("Auto select TI CC device: %s", device)
        port = device.device

    xonxoff, rtscts = False, False
    if config[CONF_FLOW_CONTROL] == "hardware":
        xonxoff, rtscts = False, True
    elif config[CONF_FLOW_CONTROL] == "software":
        xonxoff, rtscts = True, False

    LOGGER.debug("Connecting on port %s with boudrate %d", port, baudrate)
    _, protocol = await serial_asyncio.create_serial_connection(
        loop,
        lambda: protocol,
        url=port,
        baudrate=baudrate,
        parity=serial.PARITY_NONE,
        stopbits=serial.STOPBITS_ONE,
        xonxoff=xonxoff,
        rtscts=rtscts,
    )

    await connected_future

    protocol.write(b"\xef")
    await asyncio.sleep(1)

    return protocol



import asyncio
import logging
from typing import Any, Dict, Optional

import serial
import zigpy.exceptions

from zigpy_cc import uart
from zigpy_cc.config import CONF_DEVICE_PATH, SCHEMA_DEVICE
from zigpy_cc.definition import Definition
from zigpy_cc.exception import CommandError
from zigpy_cc.types import CommandType, Repr, Subsystem, Timeouts
from zigpy_cc.uart import Gateway
from zigpy_cc.zpi_object import ZpiObject

LOGGER = logging.getLogger(__name__)

COMMAND_TIMEOUT = 2


class Matcher(Repr):
    def __init__(self, command_type, subsystem, command, payload):
        self.command_type = command_type
        self.subsystem = subsystem
        self.command = command
        self.payload = payload


class Waiter(Repr):
    def __init__(
        self,
        waiter_id: int,
        command_type: int,
        subsystem: int,
        command: str,
        payload,
        timeout: int,
        sequence,
    ):
        self.id = waiter_id
        self.matcher = Matcher(command_type, subsystem, command, payload)
        self.future = asyncio.get_event_loop().create_future()
        self.timeout = timeout
        self.sequence = sequence

    async def wait(self):
        return await asyncio.wait_for(self.future, self.timeout / 1000)

    def set_result(self, result) -> None:
        if self.future.cancelled():
            LOGGER.warning("Waiter already cancelled: %s", self)
        elif self.future.done():
            LOGGER.warning("Waiter already done: %s", self)
        else:
            self.future.set_result(result)

    def match(self, obj: ZpiObject):
        matcher = self.matcher
        if (
            matcher.command_type != obj.command_type
            or matcher.subsystem != obj.subsystem
            or matcher.command != obj.command
        ):
            return False

        if matcher.payload:
            for f, v in matcher.payload.items():
                if v != obj.payload[f]:
                    return False

        return True


class API:
    _uart: Optional[Gateway]

    def __init__(self, device_config: Dict[str, Any]):
        self._config = device_config
        self._lock = asyncio.Lock()
        self._waiter_id = 0
        self._waiters: Dict[int, Waiter] = {}
        self._app = None
        self._proto_ver = None
        self._uart = None

    @property
    def protocol_version(self):
        """Protocol Version."""
        return self._proto_ver

    @classmethod
    async def new(cls, application, config: Dict[str, Any]) -> "API":
        api = cls(config)
        await api.connect()
        api.set_application(application)
        return api

    def set_application(self, app):
        self._app = app

    async def connect(self):
        assert self._uart is None
        self._uart = await uart.connect(self._config, self)

    def close(self):
        if self._uart:
            self._uart.close()
            self._uart = None

    def connection_lost(self):
        self._app.connection_lost()

    async def request(
        self, subsystem, command, payload, waiter_id=None, expected_status=None
    ):
        obj = ZpiObject.from_command(subsystem, command, payload)
        return await self.request_raw(obj, waiter_id, expected_status)

    async def request_raw(self, obj: ZpiObject, waiter_id=None, expected_status=None):
        async with self._lock:
            return await self._request_raw(obj, waiter_id, expected_status)

    async def _request_raw(self, obj: ZpiObject, waiter_id=None, expected_status=None):
        if expected_status is None:
            expected_status = [0]

        LOGGER.debug("--> %s", obj)
        frame = obj.to_unpi_frame()

        if obj.command_type == CommandType.SREQ:
            timeout = (
                20000
                if obj.command == "bdbStartCommissioning"
                or obj.command == "startupFromApp"
                else Timeouts.SREQ
            )
            waiter = self.wait_for(
                CommandType.SRSP, obj.subsystem, obj.command, {}, timeout
            )
            self._uart.send(frame)
            result = await waiter.wait()
            if (
                result
                and "status" in result.payload
                and result.payload["status"] not in expected_status
            ):
                if waiter_id is not None:
                    self._waiters.pop(waiter_id).set_result(result)

                raise CommandError(
                    result.payload["status"],
                    "SREQ '{}' failed with status '{}' (expected '{}')".format(
                        obj.command, result.payload["status"], expected_status
                    ),
                )
            else:
                return result
        elif obj.command_type == CommandType.AREQ and obj.is_reset_command():
            waiter = self.wait_for(
                CommandType.AREQ, Subsystem.SYS, "resetInd", {}, Timeouts.reset
            )
            # TODO clear queue, requests waiting for lock
            self._uart.send(frame)
            return await waiter.wait()
        else:
            if obj.command_type == CommandType.AREQ:
                self._uart.send(frame)
                return None
            else:
                LOGGER.warning("Unknown type '%s'", obj.command_type)
                raise Exception("Unknown type '{}'".format(obj.command_type))

    def create_response_waiter(self, obj: ZpiObject, sequence=None):
        waiter = self.get_response_waiter(obj, sequence)
        if waiter:
            LOGGER.debug("waiting for %d %s", sequence, obj.command)
        return waiter

    def get_response_waiter(self, obj: ZpiObject, sequence=None):
        if obj.command_type == CommandType.SREQ and obj.command.startswith(
            "dataRequest"
        ):
            payload = {
                "transid": obj.payload["transid"],
            }
            return self.wait_for(CommandType.AREQ, Subsystem.AF, "dataConfirm", payload)

        if obj.command_type == CommandType.SREQ and obj.command.endswith("Req"):
            rsp = obj.command.replace("Req", "Rsp")
            for cmd in Definition[obj.subsystem]:
                if rsp == cmd["name"]:
                    payload = {"srcaddr": obj.payload["dstaddr"]}
                    return self.wait_for(
                        CommandType.AREQ, Subsystem.ZDO, rsp, payload, sequence=sequence
                    )

        LOGGER.warning("no response cmd configured for %s", obj.command)
        return None

    def wait_for(
        self,
        command_type: CommandType,
        subsystem: Subsystem,
        command: str,
        payload=None,
        timeout=Timeouts.default,
        sequence=None,
    ):
        waiter = Waiter(
            self._waiter_id,
            command_type,
            subsystem,
            command,
            payload,
            timeout,
            sequence,
        )
        self._waiters[waiter.id] = waiter
        self._waiter_id += 1

        def callback():
            if not waiter.future.done() or waiter.future.cancelled():
                LOGGER.warning(
                    "No response for: %s %s %s %s",
                    command_type.name,
                    subsystem.name,
                    command,
                    payload,
                )
                try:
                    self._waiters.pop(waiter.id)
                except KeyError:
                    LOGGER.warning("Waiter not found: %s", waiter)

        asyncio.get_event_loop().call_later(timeout / 1000 + 0.1, callback)

        return waiter

    def data_received(self, frame):
        try:
            obj = ZpiObject.from_unpi_frame(frame)
        except Exception as e:
            LOGGER.error("Error while parsing frame: %s", frame)
            raise e

        for waiter_id in list(self._waiters):
            waiter = self._waiters.get(waiter_id)
            if waiter.match(obj):
                self._waiters.pop(waiter_id)
                waiter.set_result(obj)
                if waiter.sequence:
                    obj.sequence = waiter.sequence
                    break

        LOGGER.debug("<-- %s", obj)

        if self._app is not None:
            self._app.handle_znp(obj)

        try:
            getattr(self, "_handle_%s" % (obj.command,))(obj)
        except AttributeError:
            pass

    async def version(self):
        version = await self.request(Subsystem.SYS, "version", {})
        # todo check version
        self._proto_ver = version.payload
        return version.payload

    def _handle_version(self, data):
        LOGGER.debug("Version response: %s", data.payload)

    def _handle_getDeviceInfo(self, data):
        LOGGER.info("Device info: %s", data.payload)

    def _handle_srcRtgInd(self, data):
        pass

    @classmethod
    async def probe(cls, device_config: Dict[str, Any]) -> bool:
        """Probe port for the device presence."""
        api = cls(SCHEMA_DEVICE(device_config))
        try:
            await asyncio.wait_for(api._probe(), timeout=COMMAND_TIMEOUT)
            return True
        except (
            asyncio.TimeoutError,
            serial.SerialException,
            zigpy.exceptions.ZigbeeException,
        ) as exc:
            LOGGER.debug(
                "Unsuccessful radio probe of '%s' port",
                device_config[CONF_DEVICE_PATH],
                exc_info=exc,
            )
        finally:
            api.close()

        return False

    async def _probe(self) -> None:
        """Open port and try sending a command"""
        await self.connect()
        await self.version()



from zigpy.profiles import zha
from zigpy.types import BroadcastAddress

from zigpy_cc import uart
from zigpy_cc.buffalo import Buffalo, BuffaloOptions
from zigpy_cc.definition import Definition
from zigpy_cc.types import CommandType, ParameterType, Subsystem, AddressMode

BufferAndListTypes = [
    ParameterType.BUFFER,
    ParameterType.BUFFER8,
    ParameterType.BUFFER16,
    ParameterType.BUFFER18,
    ParameterType.BUFFER32,
    ParameterType.BUFFER42,
    ParameterType.BUFFER100,
    ParameterType.LIST_UINT16,
    ParameterType.LIST_ROUTING_TABLE,
    ParameterType.LIST_BIND_TABLE,
    ParameterType.LIST_NEIGHBOR_LQI,
    ParameterType.LIST_NETWORK,
    ParameterType.LIST_ASSOC_DEV,
    ParameterType.LIST_UINT8,
]


class ZpiObject:
    def __init__(
        self,
        command_type,
        subsystem,
        command: str,
        command_id,
        payload,
        parameters,
        sequence=None,
    ):
        self.command_type = CommandType(command_type)
        self.subsystem = Subsystem(subsystem)
        self.command = command
        self.command_id = command_id
        self.payload = payload
        self.parameters = parameters
        self.sequence = sequence

    def is_reset_command(self):
        return (self.command == "resetReq" and self.subsystem == Subsystem.SYS) or (
            self.command == "systemReset" and self.subsystem == Subsystem.SAPI
        )

    def to_unpi_frame(self):
        data = Buffalo(b"")

        for p in self.parameters:
            value = self.payload[p["name"]]
            data.write_parameter(p["parameterType"], value, {})

        return uart.UnpiFrame(
            self.command_type, self.subsystem, self.command_id, data.buffer
        )

    @classmethod
    def from_command(cls, subsystem, command, payload):
        cmd = next(c for c in Definition[subsystem] if c["name"] == command)
        parameters = (
            cmd["response"] if cmd["type"] == CommandType.SRSP else cmd["request"]
        )

        return cls(cmd["type"], subsystem, cmd["name"], cmd["ID"], payload, parameters)

    @classmethod
    def from_unpi_frame(cls, frame):
        cmd = next(
            c for c in Definition[frame.subsystem] if c["ID"] == frame.command_id
        )
        parameters = (
            cmd["response"]
            if frame.command_type == CommandType.SRSP
            else cmd["request"]
        )
        payload = cls.read_parameters(frame.data, parameters)

        return cls(
            frame.command_type,
            frame.subsystem,
            cmd["name"],
            cmd["ID"],
            payload,
            parameters,
        )

    @classmethod
    def from_cluster(
        cls,
        nwk,
        profile,
        cluster,
        src_ep,
        dst_ep,
        sequence,
        data,
        *,
        radius=30,
        addr_mode=None
    ):
        if profile == zha.PROFILE_ID:
            subsystem = Subsystem.AF
            if addr_mode is None:
                cmd = next(c for c in Definition[subsystem] if c["ID"] == 1)
            else:
                cmd = next(c for c in Definition[subsystem] if c["ID"] == 2)
        else:
            subsystem = Subsystem.ZDO
            cmd = next(c for c in Definition[subsystem] if c["ID"] == cluster)
        name = cmd["name"]
        parameters = (
            cmd["response"] if cmd["type"] == CommandType.SRSP else cmd["request"]
        )

        if name == "dataRequest":
            payload = {
                "dstaddr": int(nwk),
                "destendpoint": dst_ep,
                "srcendpoint": src_ep,
                "clusterid": cluster,
                "transid": sequence,
                "options": 0,
                "radius": radius,
                "len": len(data),
                "data": data,
            }
        elif name == "dataRequestExt":
            payload = {
                "dstaddrmode": addr_mode,
                "dstaddr": nwk,
                "destendpoint": dst_ep,
                "dstpanid": 0,
                "srcendpoint": src_ep,
                "clusterid": cluster,
                "transid": sequence,
                "options": 0,
                "radius": radius,
                "len": len(data),
                "data": data,
            }
        elif name == "mgmtPermitJoinReq":
            addrmode = (
                AddressMode.ADDR_BROADCAST
                if nwk == BroadcastAddress.ALL_ROUTERS_AND_COORDINATOR
                else AddressMode.ADDR_16BIT
            )
            payload = cls.read_parameters(
                bytes([addrmode]) + nwk.to_bytes(2, "little") + data[1:], parameters
            )
        else:
            # TODO
            # assert sequence == data[0]
            payload = cls.read_parameters(
                nwk.to_bytes(2, "little") + data[1:], parameters
            )

        return cls(
            cmd["type"], subsystem, name, cmd["ID"], payload, parameters, sequence
        )

    @classmethod
    def read_parameters(cls, data: bytes, parameters):
        # print(parameters)
        buffalo = Buffalo(data)
        res = {}
        length = None
        start_index = None
        for p in parameters:
            options = BuffaloOptions()
            name = p["name"]
            param_type = p["parameterType"]
            if param_type in BufferAndListTypes:
                if isinstance(length, int):
                    options.length = length

                if param_type == ParameterType.LIST_ASSOC_DEV:
                    if isinstance(start_index, int):
                        options.startIndex = start_index

            res[name] = buffalo.read_parameter(name, param_type, options)

            # For LIST_ASSOC_DEV, we need to grab the start_index which is
            # right before the length
            start_index = length
            # When reading a buffer, assume that the previous parsed parameter
            # contains the length of the buffer
            length = res[name]

        return res

    def __repr__(self) -> str:
        command_type = CommandType(self.command_type).name
        subsystem = Subsystem(self.subsystem).name
        return "{} {} {} tsn: {} {}".format(
            command_type, subsystem, self.command, self.sequence, self.payload
        )



"""
GENERATED BY get_definitions.py
"""
from zigpy_cc.types import Subsystem, CommandType, ParameterType

name = "name"
ID = "ID"
request = "request"
response = "response"
parameterType = "parameterType"
type = "type"


Definition = {
    Subsystem.SYS: [
        {
            name: "resetReq",
            ID: 0,
            type: CommandType.AREQ,
            request: [{name: "type", parameterType: ParameterType.UINT8}],
        },
        {
            name: "ping",
            ID: 1,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "capabilities", parameterType: ParameterType.UINT16}],
        },
        {
            name: "version",
            ID: 2,
            type: CommandType.SREQ,
            request: [],
            response: [
                {name: "transportrev", parameterType: ParameterType.UINT8},
                {name: "product", parameterType: ParameterType.UINT8},
                {name: "majorrel", parameterType: ParameterType.UINT8},
                {name: "minorrel", parameterType: ParameterType.UINT8},
                {name: "maintrel", parameterType: ParameterType.UINT8},
                {name: "revision", parameterType: ParameterType.UINT32},
            ],
        },
        {
            name: "setExtAddr",
            ID: 3,
            type: CommandType.SREQ,
            request: [{name: "extaddress", parameterType: ParameterType.IEEEADDR}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "getExtAddr",
            ID: 4,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "extaddress", parameterType: ParameterType.IEEEADDR}],
        },
        {
            name: "ramRead",
            ID: 5,
            type: CommandType.SREQ,
            request: [
                {name: "address", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT8},
            ],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "ramWrite",
            ID: 6,
            type: CommandType.SREQ,
            request: [
                {name: "address", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "osalNvItemInit",
            ID: 7,
            type: CommandType.SREQ,
            request: [
                {name: "id", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT16},
                {name: "initlen", parameterType: ParameterType.UINT8},
                {name: "initvalue", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "osalNvRead",
            ID: 8,
            type: CommandType.SREQ,
            request: [
                {name: "id", parameterType: ParameterType.UINT16},
                {name: "offset", parameterType: ParameterType.UINT8},
            ],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "osalNvWrite",
            ID: 9,
            type: CommandType.SREQ,
            request: [
                {name: "id", parameterType: ParameterType.UINT16},
                {name: "offset", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "osalStartTimer",
            ID: 10,
            type: CommandType.SREQ,
            request: [
                {name: "id", parameterType: ParameterType.UINT8},
                {name: "timeout", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "osalStopTimer",
            ID: 11,
            type: CommandType.SREQ,
            request: [{name: "id", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "random",
            ID: 12,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "value", parameterType: ParameterType.UINT16}],
        },
        {
            name: "adcRead",
            ID: 13,
            type: CommandType.SREQ,
            request: [
                {name: "channel", parameterType: ParameterType.UINT8},
                {name: "resolution", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "value", parameterType: ParameterType.UINT16}],
        },
        {
            name: "gpio",
            ID: 14,
            type: CommandType.SREQ,
            request: [
                {name: "operation", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "value", parameterType: ParameterType.UINT8}],
        },
        {
            name: "stackTune",
            ID: 15,
            type: CommandType.SREQ,
            request: [
                {name: "operation", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.INT8},
            ],
            response: [{name: "value", parameterType: ParameterType.UINT8}],
        },
        {
            name: "setTime",
            ID: 16,
            type: CommandType.SREQ,
            request: [
                {name: "utc", parameterType: ParameterType.UINT32},
                {name: "hour", parameterType: ParameterType.UINT8},
                {name: "minute", parameterType: ParameterType.UINT8},
                {name: "second", parameterType: ParameterType.UINT8},
                {name: "month", parameterType: ParameterType.UINT8},
                {name: "day", parameterType: ParameterType.UINT8},
                {name: "year", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "getTime",
            ID: 17,
            type: CommandType.SREQ,
            request: [],
            response: [
                {name: "utc", parameterType: ParameterType.UINT32},
                {name: "hour", parameterType: ParameterType.UINT8},
                {name: "minute", parameterType: ParameterType.UINT8},
                {name: "second", parameterType: ParameterType.UINT8},
                {name: "month", parameterType: ParameterType.UINT8},
                {name: "day", parameterType: ParameterType.UINT8},
                {name: "year", parameterType: ParameterType.UINT16},
            ],
        },
        {
            name: "osalNvDelete",
            ID: 18,
            type: CommandType.SREQ,
            request: [
                {name: "id", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "osalNvLength",
            ID: 19,
            type: CommandType.SREQ,
            request: [{name: "id", parameterType: ParameterType.UINT16}],
            response: [{name: "length", parameterType: ParameterType.UINT16}],
        },
        {
            name: "setTxPower",
            ID: 20,
            type: CommandType.SREQ,
            request: [{name: "level", parameterType: ParameterType.UINT8}],
            response: [{name: "txpower", parameterType: ParameterType.UINT8}],
        },
        {
            name: "jammerParameters",
            ID: 21,
            type: CommandType.SREQ,
            request: [
                {name: "jmrcntievents", parameterType: ParameterType.UINT16},
                {name: "jmrhinoiselvl", parameterType: ParameterType.UINT8},
                {name: "jmrdetectperiod", parameterType: ParameterType.UINT32},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "snifferParameters",
            ID: 22,
            type: CommandType.SREQ,
            request: [{name: "param", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "zdiagsInitStats",
            ID: 23,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "zdiagsClearStats",
            ID: 24,
            type: CommandType.SREQ,
            request: [{name: "clearnv", parameterType: ParameterType.UINT8}],
            response: [{name: "sysclock", parameterType: ParameterType.UINT32}],
        },
        {
            name: "zdiagsGetStats",
            ID: 25,
            type: CommandType.SREQ,
            request: [{name: "attributeid", parameterType: ParameterType.UINT16}],
            response: [{name: "attributevalue", parameterType: ParameterType.UINT32}],
        },
        {
            name: "zdiagsRestoreStatsNv",
            ID: 26,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "zdiagsSaveStatsToNv",
            ID: 27,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "sysclock", parameterType: ParameterType.UINT32}],
        },
        {
            name: "osalNvReadExt",
            ID: 28,
            type: CommandType.SREQ,
            request: [
                {name: "id", parameterType: ParameterType.UINT16},
                {name: "offset", parameterType: ParameterType.UINT16},
            ],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "osalNvWriteExt",
            ID: 29,
            type: CommandType.SREQ,
            request: [
                {name: "id", parameterType: ParameterType.UINT16},
                {name: "offset", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT16},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "nvCreate",
            ID: 48,
            type: CommandType.SREQ,
            request: [
                {name: "sysid", parameterType: ParameterType.UINT8},
                {name: "itemid", parameterType: ParameterType.UINT16},
                {name: "subid", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT32},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "nvDelete",
            ID: 49,
            type: CommandType.SREQ,
            request: [
                {name: "sysid", parameterType: ParameterType.UINT8},
                {name: "itemid", parameterType: ParameterType.UINT16},
                {name: "subid", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "nvLength",
            ID: 50,
            type: CommandType.SREQ,
            request: [
                {name: "sysid", parameterType: ParameterType.UINT8},
                {name: "itemid", parameterType: ParameterType.UINT16},
                {name: "subid", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "len", parameterType: ParameterType.UINT8}],
        },
        {
            name: "nvRead",
            ID: 51,
            type: CommandType.SREQ,
            request: [
                {name: "sysid", parameterType: ParameterType.UINT8},
                {name: "itemid", parameterType: ParameterType.UINT16},
                {name: "subid", parameterType: ParameterType.UINT16},
                {name: "offset", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT8},
            ],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "nvWrite",
            ID: 52,
            type: CommandType.SREQ,
            request: [
                {name: "sysid", parameterType: ParameterType.UINT8},
                {name: "itemid", parameterType: ParameterType.UINT16},
                {name: "subid", parameterType: ParameterType.UINT16},
                {name: "offset", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "nvUpdate",
            ID: 53,
            type: CommandType.SREQ,
            request: [
                {name: "sysid", parameterType: ParameterType.UINT8},
                {name: "itemid", parameterType: ParameterType.UINT16},
                {name: "subid", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "nvCompact",
            ID: 54,
            type: CommandType.SREQ,
            request: [{name: "threshold", parameterType: ParameterType.UINT16}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "resetInd",
            ID: 128,
            type: CommandType.AREQ,
            request: [
                {name: "reason", parameterType: ParameterType.UINT8},
                {name: "transportrev", parameterType: ParameterType.UINT8},
                {name: "productid", parameterType: ParameterType.UINT8},
                {name: "majorrel", parameterType: ParameterType.UINT8},
                {name: "minorrel", parameterType: ParameterType.UINT8},
                {name: "hwrev", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "osalTimerExpired",
            ID: 129,
            type: CommandType.AREQ,
            request: [{name: "id", parameterType: ParameterType.UINT8}],
        },
        {
            name: "jammerInd",
            ID: 130,
            type: CommandType.AREQ,
            request: [{name: "jammerind", parameterType: ParameterType.UINT8}],
        },
    ],
    Subsystem.MAC: [
        {
            name: "resetReq",
            ID: 1,
            type: CommandType.SREQ,
            request: [{name: "setdefault", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "init",
            ID: 2,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "startReq",
            ID: 3,
            type: CommandType.SREQ,
            request: [
                {name: "starttime", parameterType: ParameterType.UINT32},
                {name: "panid", parameterType: ParameterType.UINT16},
                {name: "logicalchannel", parameterType: ParameterType.UINT8},
                {name: "channelpage", parameterType: ParameterType.UINT8},
                {name: "beaconorder", parameterType: ParameterType.UINT8},
                {name: "superframeorder", parameterType: ParameterType.UINT8},
                {name: "pancoordinator", parameterType: ParameterType.UINT8},
                {name: "batterylifeext", parameterType: ParameterType.UINT8},
                {name: "coordrealignment", parameterType: ParameterType.UINT8},
                {name: "realignkeysource", parameterType: ParameterType.BUFFER},
                {name: "realignsecuritylevel", parameterType: ParameterType.UINT8},
                {name: "realignkeyidmode", parameterType: ParameterType.UINT8},
                {name: "realignkeyindex", parameterType: ParameterType.UINT8},
                {name: "beaconkeysource", parameterType: ParameterType.BUFFER},
                {name: "beaconsecuritylevel", parameterType: ParameterType.UINT8},
                {name: "beaconkeyidmode", parameterType: ParameterType.UINT8},
                {name: "beaconkeyindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "syncReq",
            ID: 4,
            type: CommandType.SREQ,
            request: [
                {name: "logicalchannel", parameterType: ParameterType.UINT8},
                {name: "channelpage", parameterType: ParameterType.UINT8},
                {name: "trackbeacon", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "dataReq",
            ID: 5,
            type: CommandType.SREQ,
            request: [
                {name: "destaddressmode", parameterType: ParameterType.UINT8},
                {name: "destaddress", parameterType: ParameterType.IEEEADDR},
                {name: "destpanid", parameterType: ParameterType.UINT16},
                {name: "srcaddressmode", parameterType: ParameterType.UINT8},
                {name: "handle", parameterType: ParameterType.UINT8},
                {name: "txoption", parameterType: ParameterType.UINT8},
                {name: "logicalchannel", parameterType: ParameterType.UINT8},
                {name: "power", parameterType: ParameterType.UINT8},
                {name: "keysource", parameterType: ParameterType.BUFFER},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
                {name: "msdulength", parameterType: ParameterType.UINT8},
                {name: "msdu", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "associateReq",
            ID: 6,
            type: CommandType.SREQ,
            request: [
                {name: "logicalchannel", parameterType: ParameterType.UINT8},
                {name: "channelpage", parameterType: ParameterType.UINT8},
                {name: "coordaddressmode", parameterType: ParameterType.UINT8},
                {name: "coordaddress", parameterType: ParameterType.IEEEADDR},
                {name: "coordpanid", parameterType: ParameterType.UINT16},
                {name: "capabilityinformation", parameterType: ParameterType.UINT8},
                {name: "keysource", parameterType: ParameterType.BUFFER},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "disassociateReq",
            ID: 7,
            type: CommandType.SREQ,
            request: [
                {name: "deviceaddressmode", parameterType: ParameterType.UINT8},
                {name: "deviceaddress", parameterType: ParameterType.IEEEADDR},
                {name: "devicepanid", parameterType: ParameterType.UINT16},
                {name: "disassociatereason", parameterType: ParameterType.UINT8},
                {name: "txindirect", parameterType: ParameterType.UINT8},
                {name: "keysource", parameterType: ParameterType.BUFFER},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "getReq",
            ID: 8,
            type: CommandType.SREQ,
            request: [{name: "attribute", parameterType: ParameterType.UINT8}],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "data", parameterType: ParameterType.BUFFER16},
            ],
        },
        {
            name: "setReq",
            ID: 9,
            type: CommandType.SREQ,
            request: [
                {name: "attribute", parameterType: ParameterType.UINT8},
                {name: "attributevalue", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "scanReq",
            ID: 12,
            type: CommandType.SREQ,
            request: [
                {name: "scanchannels", parameterType: ParameterType.UINT32},
                {name: "scantype", parameterType: ParameterType.UINT8},
                {name: "scanduration", parameterType: ParameterType.UINT8},
                {name: "channelpage", parameterType: ParameterType.UINT8},
                {name: "maxresults", parameterType: ParameterType.UINT8},
                {name: "keysource", parameterType: ParameterType.BUFFER},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "pollReq",
            ID: 13,
            type: CommandType.SREQ,
            request: [
                {name: "coordaddressmode", parameterType: ParameterType.UINT8},
                {name: "coordaddress", parameterType: ParameterType.IEEEADDR},
                {name: "coordpanid", parameterType: ParameterType.UINT16},
                {name: "keysource", parameterType: ParameterType.BUFFER},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "purgeReq",
            ID: 14,
            type: CommandType.SREQ,
            request: [{name: "msduhandle", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "setRxGainReq",
            ID: 15,
            type: CommandType.SREQ,
            request: [{name: "mode", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "securityGetReq",
            ID: 48,
            type: CommandType.SREQ,
            request: [
                {name: "attribute", parameterType: ParameterType.UINT8},
                {name: "index1", parameterType: ParameterType.UINT8},
                {name: "index2", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "securitySetReq",
            ID: 49,
            type: CommandType.SREQ,
            request: [
                {name: "attribute", parameterType: ParameterType.UINT8},
                {name: "attributevalue", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "associateRsp",
            ID: 80,
            type: CommandType.SREQ,
            request: [
                {name: "extaddr", parameterType: ParameterType.IEEEADDR},
                {name: "assocshortaddress", parameterType: ParameterType.UINT16},
                {name: "assocstatus", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "orphanRsp",
            ID: 81,
            type: CommandType.SREQ,
            request: [
                {name: "extaddr", parameterType: ParameterType.IEEEADDR},
                {name: "assocshortaddress", parameterType: ParameterType.UINT16},
                {name: "associatedmember", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "syncLossInd",
            ID: 128,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "panid", parameterType: ParameterType.UINT16},
                {name: "logicalchannel", parameterType: ParameterType.UINT8},
                {name: "channelpage", parameterType: ParameterType.UINT8},
                {name: "keysource", parameterType: ParameterType.BUFFER8},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "associateInd",
            ID: 129,
            type: CommandType.AREQ,
            request: [
                {name: "deviceextendedaddress", parameterType: ParameterType.IEEEADDR},
                {name: "capabilities", parameterType: ParameterType.UINT8},
                {name: "keysource", parameterType: ParameterType.BUFFER8},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "associateCnf",
            ID: 130,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "deviceshortaddress", parameterType: ParameterType.UINT16},
                {name: "keysource", parameterType: ParameterType.BUFFER8},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "beaconNotifyInd",
            ID: 131,
            type: CommandType.AREQ,
            request: [
                {name: "bsn", parameterType: ParameterType.UINT8},
                {name: "timestamp", parameterType: ParameterType.UINT32},
                {name: "coordinatoraddressmode", parameterType: ParameterType.UINT8},
                {
                    name: "coordinatorextendedaddress",
                    parameterType: ParameterType.IEEEADDR,
                },
                {name: "panid", parameterType: ParameterType.UINT16},
                {name: "superframespec", parameterType: ParameterType.UINT16},
                {name: "logicalchannel", parameterType: ParameterType.UINT8},
                {name: "gtspermit", parameterType: ParameterType.UINT8},
                {name: "linkquality", parameterType: ParameterType.UINT8},
                {name: "securityfailure", parameterType: ParameterType.UINT8},
                {name: "keysource", parameterType: ParameterType.BUFFER8},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
                {name: "pendingaddrspec", parameterType: ParameterType.UINT8},
                {name: "addresslist", parameterType: ParameterType.BUFFER32},
                {name: "sdulength", parameterType: ParameterType.UINT8},
                {name: "nsdu", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "dataCnf",
            ID: 132,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "handle", parameterType: ParameterType.UINT8},
                {name: "timestamp", parameterType: ParameterType.UINT32},
                {name: "timestamp2", parameterType: ParameterType.UINT16},
            ],
        },
        {
            name: "dataInd",
            ID: 133,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddrmode", parameterType: ParameterType.UINT8},
                {name: "srcaddr", parameterType: ParameterType.IEEEADDR},
                {name: "dstaddrmode", parameterType: ParameterType.UINT8},
                {name: "dstaddr", parameterType: ParameterType.IEEEADDR},
                {name: "timestamp", parameterType: ParameterType.UINT32},
                {name: "timestamp2", parameterType: ParameterType.UINT16},
                {name: "srcpanid", parameterType: ParameterType.UINT16},
                {name: "dstpanid", parameterType: ParameterType.UINT16},
                {name: "linkquality", parameterType: ParameterType.UINT8},
                {name: "correlation", parameterType: ParameterType.UINT8},
                {name: "rssi", parameterType: ParameterType.UINT8},
                {name: "dsn", parameterType: ParameterType.UINT8},
                {name: "keysource", parameterType: ParameterType.BUFFER8},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
                {name: "length", parameterType: ParameterType.UINT8},
                {name: "data", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "disassociateInd",
            ID: 134,
            type: CommandType.AREQ,
            request: [
                {name: "extendedaddress", parameterType: ParameterType.IEEEADDR},
                {name: "disassociatereason", parameterType: ParameterType.UINT8},
                {name: "keysource", parameterType: ParameterType.BUFFER8},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "disassociateCnf",
            ID: 135,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "deviceaddrmode", parameterType: ParameterType.UINT8},
                {name: "deviceaddr", parameterType: ParameterType.IEEEADDR},
                {name: "devicepanid", parameterType: ParameterType.UINT16},
            ],
        },
        {
            name: "orphanInd",
            ID: 138,
            type: CommandType.AREQ,
            request: [
                {name: "extendedaddr", parameterType: ParameterType.IEEEADDR},
                {name: "keysource", parameterType: ParameterType.BUFFER8},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "pollCnf",
            ID: 139,
            type: CommandType.AREQ,
            request: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "scanCnf",
            ID: 140,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "ed", parameterType: ParameterType.UINT8},
                {name: "scantype", parameterType: ParameterType.UINT8},
                {name: "channelpage", parameterType: ParameterType.UINT8},
                {name: "unscannedchannellist", parameterType: ParameterType.UINT32},
                {name: "resultlistcount", parameterType: ParameterType.UINT8},
                {name: "resultlistmaxlength", parameterType: ParameterType.UINT8},
                {name: "resultlist", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "commStatusInd",
            ID: 141,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "srcaddrmode", parameterType: ParameterType.UINT8},
                {name: "srcaddr", parameterType: ParameterType.IEEEADDR},
                {name: "dstaddrmode", parameterType: ParameterType.UINT8},
                {name: "dstaddr", parameterType: ParameterType.IEEEADDR},
                {name: "devicepanid", parameterType: ParameterType.UINT16},
                {name: "reason", parameterType: ParameterType.UINT8},
                {name: "keysource", parameterType: ParameterType.BUFFER8},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "keyidmode", parameterType: ParameterType.UINT8},
                {name: "keyindex", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "startCnf",
            ID: 142,
            type: CommandType.AREQ,
            request: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "rxEnableCnf",
            ID: 143,
            type: CommandType.AREQ,
            request: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "purgeCnf",
            ID: 144,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "handle", parameterType: ParameterType.UINT8},
            ],
        },
    ],
    Subsystem.AF: [
        {
            name: "register",
            ID: 0,
            type: CommandType.SREQ,
            request: [
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "appprofid", parameterType: ParameterType.UINT16},
                {name: "appdeviceid", parameterType: ParameterType.UINT16},
                {name: "appdevver", parameterType: ParameterType.UINT8},
                {name: "latencyreq", parameterType: ParameterType.UINT8},
                {name: "appnuminclusters", parameterType: ParameterType.UINT8},
                {name: "appinclusterlist", parameterType: ParameterType.LIST_UINT16},
                {name: "appnumoutclusters", parameterType: ParameterType.UINT8},
                {name: "appoutclusterlist", parameterType: ParameterType.LIST_UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "dataRequest",
            ID: 1,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "destendpoint", parameterType: ParameterType.UINT8},
                {name: "srcendpoint", parameterType: ParameterType.UINT8},
                {name: "clusterid", parameterType: ParameterType.UINT16},
                {name: "transid", parameterType: ParameterType.UINT8},
                {name: "options", parameterType: ParameterType.UINT8},
                {name: "radius", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "data", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "dataRequestExt",
            ID: 2,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddrmode", parameterType: ParameterType.UINT8},
                {name: "dstaddr", parameterType: ParameterType.IEEEADDR},
                {name: "destendpoint", parameterType: ParameterType.UINT8},
                {name: "dstpanid", parameterType: ParameterType.UINT16},
                {name: "srcendpoint", parameterType: ParameterType.UINT8},
                {name: "clusterid", parameterType: ParameterType.UINT16},
                {name: "transid", parameterType: ParameterType.UINT8},
                {name: "options", parameterType: ParameterType.UINT8},
                {name: "radius", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT16},
                {name: "data", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "dataRequestSrcRtg",
            ID: 3,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "destendpoint", parameterType: ParameterType.UINT8},
                {name: "srcendpoint", parameterType: ParameterType.UINT8},
                {name: "clusterid", parameterType: ParameterType.UINT16},
                {name: "transid", parameterType: ParameterType.UINT8},
                {name: "options", parameterType: ParameterType.UINT8},
                {name: "radius", parameterType: ParameterType.UINT8},
                {name: "relaycount", parameterType: ParameterType.UINT8},
                {name: "relaylist", parameterType: ParameterType.LIST_UINT16},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "data", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "delete",
            ID: 4,
            type: CommandType.SREQ,
            request: [{name: "endpoint", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "interPanCtl",
            ID: 16,
            type: CommandType.SREQ,
            request: [
                {name: "cmd", parameterType: ParameterType.UINT8},
                {name: "data", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "dataStore",
            ID: 17,
            type: CommandType.SREQ,
            request: [
                {name: "index", parameterType: ParameterType.UINT16},
                {name: "length", parameterType: ParameterType.UINT8},
                {name: "data", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "dataRetrieve",
            ID: 18,
            type: CommandType.SREQ,
            request: [
                {name: "timestamp", parameterType: ParameterType.UINT32},
                {name: "index", parameterType: ParameterType.UINT16},
                {name: "length", parameterType: ParameterType.UINT8},
            ],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "length", parameterType: ParameterType.UINT8},
                {name: "data", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "apsfConfigSet",
            ID: 19,
            type: CommandType.SREQ,
            request: [
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "framedelay", parameterType: ParameterType.UINT8},
                {name: "windowsize", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "apsfConfigGet",
            ID: 20,
            type: CommandType.SREQ,
            request: [{name: "endpoint", parameterType: ParameterType.UINT8}],
            response: [
                {name: "framedelay", parameterType: ParameterType.UINT8},
                {name: "windowsize", parameterType: ParameterType.UINT8},
                {name: "nomean", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "dataConfirm",
            ID: 128,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "transid", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "incomingMsg",
            ID: 129,
            type: CommandType.AREQ,
            request: [
                {name: "groupid", parameterType: ParameterType.UINT16},
                {name: "clusterid", parameterType: ParameterType.UINT16},
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "srcendpoint", parameterType: ParameterType.UINT8},
                {name: "dstendpoint", parameterType: ParameterType.UINT8},
                {name: "wasbroadcast", parameterType: ParameterType.UINT8},
                {name: "linkquality", parameterType: ParameterType.UINT8},
                {name: "securityuse", parameterType: ParameterType.UINT8},
                {name: "timestamp", parameterType: ParameterType.UINT32},
                {name: "transseqnumber", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "data", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "incomingMsgExt",
            ID: 130,
            type: CommandType.AREQ,
            request: [
                {name: "groupid", parameterType: ParameterType.UINT16},
                {name: "clusterid", parameterType: ParameterType.UINT16},
                {name: "srcaddrmode", parameterType: ParameterType.UINT8},
                {name: "srcaddr", parameterType: ParameterType.IEEEADDR},
                {name: "srcendpoint", parameterType: ParameterType.UINT8},
                {name: "srcpanid", parameterType: ParameterType.UINT16},
                {name: "dstendpoint", parameterType: ParameterType.UINT8},
                {name: "wasbroadcast", parameterType: ParameterType.UINT8},
                {name: "linkquality", parameterType: ParameterType.UINT8},
                {name: "securityuse", parameterType: ParameterType.UINT8},
                {name: "timestamp", parameterType: ParameterType.UINT32},
                {name: "transseqnumber", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT16},
                {name: "data", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "reflectError",
            ID: 131,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "transid", parameterType: ParameterType.UINT8},
                {name: "dstaddrmode", parameterType: ParameterType.UINT8},
                {name: "dstaddr", parameterType: ParameterType.UINT16},
            ],
        },
    ],
    Subsystem.ZDO: [
        {
            name: "nwkAddrReq",
            ID: 0,
            type: CommandType.SREQ,
            request: [
                {name: "ieeeaddr", parameterType: ParameterType.IEEEADDR},
                {name: "reqtype", parameterType: ParameterType.UINT8},
                {name: "startindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "ieeeAddrReq",
            ID: 1,
            type: CommandType.SREQ,
            request: [
                {name: "shortaddr", parameterType: ParameterType.UINT16},
                {name: "reqtype", parameterType: ParameterType.UINT8},
                {name: "startindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "nodeDescReq",
            ID: 2,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddrofinterest", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "powerDescReq",
            ID: 3,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddrofinterest", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "simpleDescReq",
            ID: 4,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddrofinterest", parameterType: ParameterType.UINT16},
                {name: "endpoint", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "activeEpReq",
            ID: 5,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddrofinterest", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "matchDescReq",
            ID: 6,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddrofinterest", parameterType: ParameterType.UINT16},
                {name: "profileid", parameterType: ParameterType.UINT16},
                {name: "numinclusters", parameterType: ParameterType.UINT8},
                {name: "inclusterlist", parameterType: ParameterType.LIST_UINT16},
                {name: "numoutclusters", parameterType: ParameterType.UINT8},
                {name: "outclusterlist", parameterType: ParameterType.LIST_UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "complexDescReq",
            ID: 7,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddrofinterest", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "userDescReq",
            ID: 8,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddrofinterest", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "endDeviceAnnce",
            ID: 10,
            type: CommandType.SREQ,
            request: [
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "ieeeaddr", parameterType: ParameterType.IEEEADDR},
                {name: "capability", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "userDescSet",
            ID: 11,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddrofinterest", parameterType: ParameterType.UINT16},
                {name: "descriptor_len", parameterType: ParameterType.UINT8},
                {name: "userdescriptor", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "serverDiscReq",
            ID: 12,
            type: CommandType.SREQ,
            request: [{name: "servermask", parameterType: ParameterType.UINT16}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "endDeviceBindReq",
            ID: 32,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "localcoord", parameterType: ParameterType.UINT16},
                {name: "localieee", parameterType: ParameterType.IEEEADDR},
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "profileid", parameterType: ParameterType.UINT16},
                {name: "numinclusters", parameterType: ParameterType.UINT8},
                {name: "inclusterlist", parameterType: ParameterType.LIST_UINT16},
                {name: "numoutclusters", parameterType: ParameterType.UINT8},
                {name: "outclusterlist", parameterType: ParameterType.LIST_UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "bindReq",
            ID: 33,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "srcaddr", parameterType: ParameterType.IEEEADDR},
                {name: "srcendpoint", parameterType: ParameterType.UINT8},
                {name: "clusterid", parameterType: ParameterType.UINT16},
                {name: "dstaddrmode", parameterType: ParameterType.UINT8},
                {name: "dstaddress", parameterType: ParameterType.IEEEADDR},
                {name: "dstendpoint", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "unbindReq",
            ID: 34,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "srcaddr", parameterType: ParameterType.IEEEADDR},
                {name: "srcendpoint", parameterType: ParameterType.UINT8},
                {name: "clusterid", parameterType: ParameterType.UINT16},
                {name: "dstaddrmode", parameterType: ParameterType.UINT8},
                {name: "dstaddress", parameterType: ParameterType.IEEEADDR},
                {name: "dstendpoint", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "setLinkKey",
            ID: 35,
            type: CommandType.SREQ,
            request: [
                {name: "shortaddr", parameterType: ParameterType.UINT16},
                {name: "ieeeaddr", parameterType: ParameterType.IEEEADDR},
                {name: "linkkey", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "removeLinkKey",
            ID: 36,
            type: CommandType.SREQ,
            request: [{name: "ieeeaddr", parameterType: ParameterType.IEEEADDR}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "getLinkKey",
            ID: 37,
            type: CommandType.SREQ,
            request: [{name: "ieeeaddr", parameterType: ParameterType.IEEEADDR}],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "ieeeaddr", parameterType: ParameterType.IEEEADDR},
                {name: "linkkeydata", parameterType: ParameterType.BUFFER16},
            ],
        },
        {
            name: "nwkDiscoveryReq",
            ID: 38,
            type: CommandType.SREQ,
            request: [
                {name: "scanchannels", parameterType: ParameterType.UINT32},
                {name: "scanduration", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "joinReq",
            ID: 39,
            type: CommandType.SREQ,
            request: [
                {name: "logicalchannel", parameterType: ParameterType.UINT8},
                {name: "panid", parameterType: ParameterType.UINT16},
                {name: "extendedpanid", parameterType: ParameterType.IEEEADDR},
                {name: "chosenparent", parameterType: ParameterType.UINT16},
                {name: "parentdepth", parameterType: ParameterType.UINT8},
                {name: "stackprofile", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "mgmtNwkDiscReq",
            ID: 48,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "scanchannels", parameterType: ParameterType.UINT32},
                {name: "scanduration", parameterType: ParameterType.UINT8},
                {name: "startindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "mgmtLqiReq",
            ID: 49,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "startindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "mgmtRtgReq",
            ID: 50,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "startindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "mgmtBindReq",
            ID: 51,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "startindex", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "mgmtLeaveReq",
            ID: 52,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "deviceaddress", parameterType: ParameterType.IEEEADDR},
                {name: "removechildrenRejoin", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "mgmtDirectJoinReq",
            ID: 53,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "deviceaddr", parameterType: ParameterType.IEEEADDR},
                {name: "capinfo", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "mgmtPermitJoinReq",
            ID: 54,
            type: CommandType.SREQ,
            request: [
                {name: "addrmode", parameterType: ParameterType.UINT8},
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "duration", parameterType: ParameterType.UINT8},
                {name: "tcsignificance", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "mgmtNwkUpdateReq",
            ID: 55,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "dstaddrmode", parameterType: ParameterType.UINT8},
                {name: "channelmask", parameterType: ParameterType.UINT32},
                {name: "scanduration", parameterType: ParameterType.UINT8},
                {name: "scancount", parameterType: ParameterType.UINT8},
                {name: "nwkmanageraddr", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "msgCbRegister",
            ID: 62,
            type: CommandType.SREQ,
            request: [{name: "clusterid", parameterType: ParameterType.UINT16}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "msgCbRemove",
            ID: 63,
            type: CommandType.SREQ,
            request: [{name: "clusterid", parameterType: ParameterType.UINT16}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "startupFromApp",
            ID: 64,
            type: CommandType.SREQ,
            request: [{name: "startdelay", parameterType: ParameterType.UINT16}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "autoFindDestination",
            ID: 65,
            type: CommandType.AREQ,
            request: [{name: "endpoint", parameterType: ParameterType.UINT8}],
        },
        {
            name: "nwkAddrRsp",
            ID: 128,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "ieeeaddr", parameterType: ParameterType.IEEEADDR},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "startindex", parameterType: ParameterType.UINT8},
                {name: "numassocdev", parameterType: ParameterType.UINT8},
                {name: "assocdevlist", parameterType: ParameterType.LIST_ASSOC_DEV},
            ],
        },
        {
            name: "ieeeAddrRsp",
            ID: 129,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "ieeeaddr", parameterType: ParameterType.IEEEADDR},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "startindex", parameterType: ParameterType.UINT8},
                {name: "numassocdev", parameterType: ParameterType.UINT8},
                {name: "assocdevlist", parameterType: ParameterType.LIST_ASSOC_DEV},
            ],
        },
        {
            name: "nodeDescRsp",
            ID: 130,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {
                    name: "logicaltype_cmplxdescavai_userdescavai",
                    parameterType: ParameterType.UINT8,
                },
                {name: "apsflags_freqband", parameterType: ParameterType.UINT8},
                {name: "maccapflags", parameterType: ParameterType.UINT8},
                {name: "manufacturercode", parameterType: ParameterType.UINT16},
                {name: "maxbuffersize", parameterType: ParameterType.UINT8},
                {name: "maxintransfersize", parameterType: ParameterType.UINT16},
                {name: "servermask", parameterType: ParameterType.UINT16},
                {name: "maxouttransfersize", parameterType: ParameterType.UINT16},
                {name: "descriptorcap", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "powerDescRsp",
            ID: 131,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {
                    name: "currentpowermode_avaipowersrc",
                    parameterType: ParameterType.UINT8,
                },
                {
                    name: "currentpowersrc_currentpowersrclevel",
                    parameterType: ParameterType.UINT8,
                },
            ],
        },
        {
            name: "simpleDescRsp",
            ID: 132,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "profileid", parameterType: ParameterType.UINT16},
                {name: "deviceid", parameterType: ParameterType.UINT16},
                {name: "deviceversion", parameterType: ParameterType.UINT8},
                {name: "numinclusters", parameterType: ParameterType.UINT8},
                {name: "inclusterlist", parameterType: ParameterType.LIST_UINT16},
                {name: "numoutclusters", parameterType: ParameterType.UINT8},
                {name: "outclusterlist", parameterType: ParameterType.LIST_UINT16},
            ],
        },
        {
            name: "activeEpRsp",
            ID: 133,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "activeepcount", parameterType: ParameterType.UINT8},
                {name: "activeeplist", parameterType: ParameterType.LIST_UINT8},
            ],
        },
        {
            name: "matchDescRsp",
            ID: 134,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "matchlength", parameterType: ParameterType.UINT8},
                {name: "matchlist", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "complexDescRsp",
            ID: 135,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "complexlength", parameterType: ParameterType.UINT8},
                {name: "complexdesclist", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "userDescRsp",
            ID: 136,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "userlength", parameterType: ParameterType.UINT8},
                {name: "userdescriptor", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "userDescConf",
            ID: 137,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
            ],
        },
        {
            name: "serverDiscRsp",
            ID: 138,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "servermask", parameterType: ParameterType.UINT16},
            ],
        },
        {
            name: "endDeviceBindRsp",
            ID: 160,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "bindRsp",
            ID: 161,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "unbindRsp",
            ID: 162,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "mgmtNwkDiscRsp",
            ID: 176,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "networkcount", parameterType: ParameterType.UINT8},
                {name: "startindex", parameterType: ParameterType.UINT8},
                {name: "networklistcount", parameterType: ParameterType.UINT8},
                {name: "networklist", parameterType: ParameterType.LIST_NETWORK},
            ],
        },
        {
            name: "mgmtLqiRsp",
            ID: 177,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "neighbortableentries", parameterType: ParameterType.UINT8},
                {name: "startindex", parameterType: ParameterType.UINT8},
                {name: "neighborlqilistcount", parameterType: ParameterType.UINT8},
                {
                    name: "neighborlqilist",
                    parameterType: ParameterType.LIST_NEIGHBOR_LQI,
                },
            ],
        },
        {
            name: "mgmtRtgRsp",
            ID: 178,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "routingtableentries", parameterType: ParameterType.UINT8},
                {name: "startindex", parameterType: ParameterType.UINT8},
                {name: "routingtablelistcount", parameterType: ParameterType.UINT8},
                {
                    name: "routingtablelist",
                    parameterType: ParameterType.LIST_ROUTING_TABLE,
                },
            ],
        },
        {
            name: "mgmtBindRsp",
            ID: 179,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "bindingtableentries", parameterType: ParameterType.UINT8},
                {name: "startindex", parameterType: ParameterType.UINT8},
                {name: "bindingtablelistcount", parameterType: ParameterType.UINT8},
                {
                    name: "bindingtablelist",
                    parameterType: ParameterType.LIST_BIND_TABLE,
                },
            ],
        },
        {
            name: "mgmtLeaveRsp",
            ID: 180,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "mgmtDirectJoinRsp",
            ID: 181,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "mgmtNwkUpdateNotify",
            ID: 184,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "scanchannels", parameterType: ParameterType.UINT32},
                {name: "totaltransmissions", parameterType: ParameterType.UINT16},
                {name: "transmissionfailures", parameterType: ParameterType.UINT16},
                {name: "channelcount", parameterType: ParameterType.UINT8},
                {name: "energyvalues", parameterType: ParameterType.LIST_UINT8},
            ],
        },
        {
            name: "mgmtPermitJoinRsp",
            ID: 182,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "stateChangeInd",
            ID: 192,
            type: CommandType.AREQ,
            request: [{name: "state", parameterType: ParameterType.UINT8}],
        },
        {
            name: "endDeviceAnnceInd",
            ID: 193,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "ieeeaddr", parameterType: ParameterType.IEEEADDR},
                {name: "capabilities", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "matchDescRspSent",
            ID: 194,
            type: CommandType.AREQ,
            request: [
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "numinclusters", parameterType: ParameterType.UINT8},
                {name: "inclusterlist", parameterType: ParameterType.LIST_UINT16},
                {name: "numoutclusters", parameterType: ParameterType.UINT8},
                {name: "outclusterlist", parameterType: ParameterType.LIST_UINT16},
            ],
        },
        {
            name: "statusErrorRsp",
            ID: 195,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "srcRtgInd",
            ID: 196,
            type: CommandType.AREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "relaycount", parameterType: ParameterType.UINT8},
                {name: "relaylist", parameterType: ParameterType.LIST_UINT16},
            ],
        },
        {
            name: "beacon_notify_ind",
            ID: 197,
            type: CommandType.AREQ,
            request: [
                {name: "beaconcount", parameterType: ParameterType.UINT8},
                {name: "beaconlist", parameterType: ParameterType.BUFFER},  # FIXME
            ],
        },
        {
            name: "joinCnf",
            ID: 198,
            type: CommandType.AREQ,
            request: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "deviceaddress", parameterType: ParameterType.UINT16},
                {name: "parentaddress", parameterType: ParameterType.UINT16},
            ],
        },
        {
            name: "nwkDiscoveryCnf",
            ID: 199,
            type: CommandType.AREQ,
            request: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "concentratorIndCb",
            ID: 200,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "extaddr", parameterType: ParameterType.IEEEADDR},
                {name: "pktCost", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "leaveInd",
            ID: 201,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "extaddr", parameterType: ParameterType.IEEEADDR},
                {name: "request", parameterType: ParameterType.UINT8},
                {name: "removechildren", parameterType: ParameterType.UINT8},
                {name: "rejoin", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "setRejoinParametersReq",
            ID: 204,
            type: CommandType.SREQ,
            request: [
                {name: "backoffduration", parameterType: ParameterType.UINT32},
                {name: "scanduration", parameterType: ParameterType.UINT32},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "msgCbIncoming",
            ID: 255,
            type: CommandType.AREQ,
            request: [
                {name: "srcaddr", parameterType: ParameterType.UINT16},
                {name: "wasbroadcast", parameterType: ParameterType.UINT8},
                {name: "clusterid", parameterType: ParameterType.UINT16},
                {name: "securityuse", parameterType: ParameterType.UINT8},
                {name: "seqnum", parameterType: ParameterType.UINT8},
                {name: "macdstaddr", parameterType: ParameterType.UINT16},
                {name: "msgdata", parameterType: ParameterType.BUFFER},  # FIXME
            ],
        },
        {
            name: "endDeviceTimeoutReq",
            ID: 13,
            type: CommandType.SREQ,
            request: [
                {name: "parentaddr", parameterType: ParameterType.UINT16},
                {name: "reqrimeout", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "sendData",
            ID: 40,
            type: CommandType.SREQ,
            request: [
                {name: "shortaddr", parameterType: ParameterType.UINT16},
                {name: "transseq", parameterType: ParameterType.UINT8},
                {name: "cmd", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "buf", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "nwkAddrOfInterestReq",
            ID: 41,
            type: CommandType.SREQ,
            request: [
                {name: "shortaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "cmd", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "secAddLinkKey",
            ID: 66,
            type: CommandType.SREQ,
            request: [
                {name: "shortaddr", parameterType: ParameterType.UINT16},
                {name: "extaddr", parameterType: ParameterType.IEEEADDR},
                {name: "linkkey", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "secEntryLookupExt",
            ID: 67,
            type: CommandType.SREQ,
            request: [{name: "extaddr", parameterType: ParameterType.IEEEADDR}],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "ami", parameterType: ParameterType.UINT16},
                {name: "keynvid", parameterType: ParameterType.UINT16},
                {name: "authenticateoption", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "secDeviceRemove",
            ID: 68,
            type: CommandType.SREQ,
            request: [{name: "extaddr", parameterType: ParameterType.IEEEADDR}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "extRouteDisc",
            ID: 69,
            type: CommandType.SREQ,
            request: [
                {name: "dstAddr", parameterType: ParameterType.UINT16},
                {name: "options", parameterType: ParameterType.UINT8},
                {name: "radius", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "extRouteCheck",
            ID: 70,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "rtstatus", parameterType: ParameterType.UINT8},
                {name: "options", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "extRemoveGroup",
            ID: 71,
            type: CommandType.SREQ,
            request: [
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "groupid", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "extRemoveAllGroup",
            ID: 72,
            type: CommandType.SREQ,
            request: [{name: "endpoint", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "extFindAllGroupsEndpoint",
            ID: 73,
            type: CommandType.SREQ,
            request: [{name: "endpoint", parameterType: ParameterType.UINT8}],
            response: [
                {name: "groups", parameterType: ParameterType.UINT8},
                {name: "grouplist", parameterType: ParameterType.LIST_UINT16},
            ],
        },
        {
            name: "extFindGroup",
            ID: 74,
            type: CommandType.SREQ,
            request: [
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "groupid", parameterType: ParameterType.UINT16},
            ],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "groupid", parameterType: ParameterType.UINT16},
                {name: "namelen", parameterType: ParameterType.UINT8},
                {name: "groupname", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "extAddGroup",
            ID: 75,
            type: CommandType.SREQ,
            request: [
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "groupid", parameterType: ParameterType.UINT16},
                {name: "namelen", parameterType: ParameterType.UINT8},
                {name: "groupname", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "extCountAllGroups",
            ID: 76,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "extRxIdle",
            ID: 77,
            type: CommandType.SREQ,
            request: [
                {name: "setflag", parameterType: ParameterType.UINT8},
                {name: "setvalue", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "extUpdateNwkKey",
            ID: 78,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "keyseqnum", parameterType: ParameterType.UINT8},
                {name: "key", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "extSwitchNwkKey",
            ID: 79,
            type: CommandType.SREQ,
            request: [
                {name: "dstaddr", parameterType: ParameterType.UINT16},
                {name: "keyseqnum", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "extNwkInfo",
            ID: 80,
            type: CommandType.SREQ,
            request: [],
            response: [
                {name: "shortaddr", parameterType: ParameterType.UINT16},
                {name: "devstate", parameterType: ParameterType.UINT8},
                {name: "panid", parameterType: ParameterType.UINT16},
                {name: "parentaddr", parameterType: ParameterType.UINT16},
                {name: "extendedpanid", parameterType: ParameterType.IEEEADDR},
                {name: "parentextaddr", parameterType: ParameterType.IEEEADDR},
                {name: "channel", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "extSecApsRemoveReq",
            ID: 81,
            type: CommandType.SREQ,
            request: [
                {name: "parentaddr", parameterType: ParameterType.UINT16},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "extaddr", parameterType: ParameterType.IEEEADDR},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "forceConcentratorChange",
            ID: 82,
            type: CommandType.SREQ,
            request: [],
            response: [],
        },
        {
            name: "extSetParams",
            ID: 83,
            type: CommandType.SREQ,
            request: [{name: "usemulticast", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "tcDeviceInd",
            ID: 202,
            type: CommandType.AREQ,
            request: [
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "extaddr", parameterType: ParameterType.IEEEADDR},
                {name: "parentaddr", parameterType: ParameterType.UINT16},
            ],
        },
        {
            name: "permitJoinInd",
            ID: 203,
            type: CommandType.AREQ,
            request: [{name: "duration", parameterType: ParameterType.UINT8}],
        },
    ],
    Subsystem.SAPI: [
        {name: "systemReset", ID: 9, type: CommandType.AREQ, request: []},
        {
            name: "startRequest",
            ID: 0,
            type: CommandType.SREQ,
            request: [],
            response: [],
        },
        {
            name: "bindDevice",
            ID: 1,
            type: CommandType.SREQ,
            request: [
                {name: "action", parameterType: ParameterType.UINT8},
                {name: "commandid", parameterType: ParameterType.UINT16},
                {name: "destination", parameterType: ParameterType.IEEEADDR},
            ],
            response: [],
        },
        {
            name: "allowBind",
            ID: 2,
            type: CommandType.SREQ,
            request: [{name: "timeout", parameterType: ParameterType.UINT8}],
            response: [],
        },
        {
            name: "sendDataRequest",
            ID: 3,
            type: CommandType.SREQ,
            request: [
                {name: "destination", parameterType: ParameterType.UINT16},
                {name: "commandid", parameterType: ParameterType.UINT16},
                {name: "handle", parameterType: ParameterType.UINT8},
                {name: "txoptions", parameterType: ParameterType.UINT8},
                {name: "radius", parameterType: ParameterType.UINT8},
                {name: "payloadlen", parameterType: ParameterType.UINT8},
                {name: "payloadvalue", parameterType: ParameterType.BUFFER},
            ],
            response: [],
        },
        {
            name: "readConfiguration",
            ID: 4,
            type: CommandType.SREQ,
            request: [{name: "configid", parameterType: ParameterType.UINT8}],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "configid", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
        },
        {
            name: "writeConfiguration",
            ID: 5,
            type: CommandType.SREQ,
            request: [
                {name: "configid", parameterType: ParameterType.UINT8},
                {name: "len", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "getDeviceInfo",
            ID: 6,
            type: CommandType.SREQ,
            request: [{name: "param", parameterType: ParameterType.UINT8}],
            response: [
                {name: "param", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.BUFFER8},
            ],
        },
        {
            name: "findDeviceRequest",
            ID: 7,
            type: CommandType.SREQ,
            request: [{name: "searchKey", parameterType: ParameterType.IEEEADDR}],
            response: [],
        },
        {
            name: "permitJoiningRequest",
            ID: 8,
            type: CommandType.SREQ,
            request: [
                {name: "destination", parameterType: ParameterType.UINT16},
                {name: "timeout", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "startConfirm",
            ID: 128,
            type: CommandType.AREQ,
            request: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "bindConfirm",
            ID: 129,
            type: CommandType.AREQ,
            request: [
                {name: "commandid", parameterType: ParameterType.UINT16},
                {name: "status", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "allowBindConfirm",
            ID: 130,
            type: CommandType.AREQ,
            request: [{name: "source", parameterType: ParameterType.UINT16}],
        },
        {
            name: "sendDataConfirm",
            ID: 131,
            type: CommandType.AREQ,
            request: [
                {name: "handle", parameterType: ParameterType.UINT8},
                {name: "status", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "findDeviceConfirm",
            ID: 133,
            type: CommandType.AREQ,
            request: [
                {name: "searchtype", parameterType: ParameterType.UINT8},
                {name: "searchkey", parameterType: ParameterType.UINT16},
                {name: "result", parameterType: ParameterType.IEEEADDR},
            ],
        },
        {
            name: "receiveDataIndication",
            ID: 135,
            type: CommandType.AREQ,
            request: [
                {name: "source", parameterType: ParameterType.UINT16},
                {name: "command", parameterType: ParameterType.UINT16},
                {name: "len", parameterType: ParameterType.UINT16},
                {name: "data", parameterType: ParameterType.BUFFER},
            ],
        },
    ],
    Subsystem.UTIL: [
        {
            name: "getDeviceInfo",
            ID: 0,
            type: CommandType.SREQ,
            request: [],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "ieeeaddr", parameterType: ParameterType.IEEEADDR},
                {name: "shortaddr", parameterType: ParameterType.UINT16},
                {name: "devicetype", parameterType: ParameterType.UINT8},
                {name: "devicestate", parameterType: ParameterType.UINT8},
                {name: "numassocdevices", parameterType: ParameterType.UINT8},
                {name: "assocdeviceslist", parameterType: ParameterType.LIST_UINT16},
            ],
        },
        {
            name: "getNvInfo",
            ID: 1,
            type: CommandType.SREQ,
            request: [],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "ieeeaddr", parameterType: ParameterType.IEEEADDR},
                {name: "scanchannels", parameterType: ParameterType.UINT32},
                {name: "panid", parameterType: ParameterType.UINT16},
                {name: "securitylevel", parameterType: ParameterType.UINT8},
                {name: "preconfigkey", parameterType: ParameterType.BUFFER16},
            ],
        },
        {
            name: "setPanid",
            ID: 2,
            type: CommandType.SREQ,
            request: [{name: "panid", parameterType: ParameterType.UINT16}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "setChannels",
            ID: 3,
            type: CommandType.SREQ,
            request: [{name: "channels", parameterType: ParameterType.UINT32}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "setSeclevel",
            ID: 4,
            type: CommandType.SREQ,
            request: [{name: "securitylevel", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "setPrecfgkey",
            ID: 5,
            type: CommandType.SREQ,
            request: [{name: "preconfigkey", parameterType: ParameterType.BUFFER}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "callbackSubCmd",
            ID: 6,
            type: CommandType.SREQ,
            request: [
                {name: "subsystemid", parameterType: ParameterType.UINT16},
                {name: "action", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "keyEvent",
            ID: 7,
            type: CommandType.SREQ,
            request: [
                {name: "keys", parameterType: ParameterType.UINT8},
                {name: "shift", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "timeAlive",
            ID: 9,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "seconds", parameterType: ParameterType.UINT32}],
        },
        {
            name: "ledControl",
            ID: 10,
            type: CommandType.SREQ,
            request: [
                {name: "ledid", parameterType: ParameterType.UINT8},
                {name: "mode", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "testLoopback",
            ID: 16,
            type: CommandType.SREQ,
            request: [{name: "data", parameterType: ParameterType.BUFFER}],
            response: [{name: "data", parameterType: ParameterType.BUFFER}],
        },
        {
            name: "dataReq",
            ID: 17,
            type: CommandType.SREQ,
            request: [{name: "securityuse", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "srcMatchEnable",
            ID: 32,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "srcMatchAddEntry",
            ID: 33,
            type: CommandType.SREQ,
            request: [
                {name: "addressmode", parameterType: ParameterType.UINT8},
                {name: "address", parameterType: ParameterType.IEEEADDR},
                {name: "panid", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "srcMatchDelEntry",
            ID: 34,
            type: CommandType.SREQ,
            request: [
                {name: "addressmode", parameterType: ParameterType.UINT8},
                {name: "address", parameterType: ParameterType.IEEEADDR},
                {name: "panid", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "srcMatchCheckSrcAddr",
            ID: 35,
            type: CommandType.SREQ,
            request: [
                {name: "addressmode", parameterType: ParameterType.UINT8},
                {name: "address", parameterType: ParameterType.IEEEADDR},
                {name: "panid", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "srcMatchAckAllPending",
            ID: 36,
            type: CommandType.SREQ,
            request: [{name: "option", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "srcMatchCheckAllPending",
            ID: 37,
            type: CommandType.SREQ,
            request: [],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "addrmgrExtAddrLookup",
            ID: 64,
            type: CommandType.SREQ,
            request: [{name: "extaddr", parameterType: ParameterType.IEEEADDR}],
            response: [{name: "nwkaddr", parameterType: ParameterType.UINT16}],
        },
        {
            name: "addrmgrNwkAddrLookup",
            ID: 65,
            type: CommandType.SREQ,
            request: [{name: "nwkaddr", parameterType: ParameterType.UINT16}],
            response: [{name: "extaddr", parameterType: ParameterType.IEEEADDR}],
        },
        {
            name: "apsmeLinkKeyDataGet",
            ID: 68,
            type: CommandType.SREQ,
            request: [{name: "extaddr", parameterType: ParameterType.IEEEADDR}],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "seckey", parameterType: ParameterType.BUFFER16},
                {name: "txfrmcntr", parameterType: ParameterType.UINT32},
                {name: "rxfrmcntr", parameterType: ParameterType.UINT32},
            ],
        },
        {
            name: "apsmeLinkKeyNvIdGet",
            ID: 69,
            type: CommandType.SREQ,
            request: [{name: "extaddr", parameterType: ParameterType.IEEEADDR}],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "linkkeynvid", parameterType: ParameterType.UINT16},
            ],
        },
        {
            name: "assocCount",
            ID: 72,
            type: CommandType.SREQ,
            request: [
                {name: "startrelation", parameterType: ParameterType.UINT8},
                {name: "endrelation", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "count", parameterType: ParameterType.UINT16}],
        },
        {
            name: "assocFindDevice",
            ID: 73,
            type: CommandType.SREQ,
            request: [{name: "number", parameterType: ParameterType.UINT8}],
            response: [{name: "device", parameterType: ParameterType.BUFFER18}],
        },
        {
            name: "assocGetWithAddress",
            ID: 74,
            type: CommandType.SREQ,
            request: [
                {name: "extaddr", parameterType: ParameterType.IEEEADDR},
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "device", parameterType: ParameterType.BUFFER18}],
        },
        {
            name: "apsmeRequestKeyCmd",
            ID: 75,
            type: CommandType.SREQ,
            request: [{name: "partneraddr", parameterType: ParameterType.IEEEADDR}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "zclKeyEstInitEst",
            ID: 128,
            type: CommandType.SREQ,
            request: [
                {name: "taskid", parameterType: ParameterType.UINT8},
                {name: "seqnum", parameterType: ParameterType.UINT8},
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "addrmode", parameterType: ParameterType.UINT8},
                {name: "extaddr", parameterType: ParameterType.IEEEADDR},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "zclKeyEstSign",
            ID: 129,
            type: CommandType.SREQ,
            request: [
                {name: "inputlen", parameterType: ParameterType.UINT8},
                {name: "input", parameterType: ParameterType.BUFFER},
            ],
            response: [
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "key", parameterType: ParameterType.BUFFER42},
            ],
        },
        {name: "syncReq", ID: 224, type: CommandType.AREQ, request: []},
        {
            name: "zclKeyEstablishInd",
            ID: 225,
            type: CommandType.AREQ,
            request: [
                {name: "taskid", parameterType: ParameterType.UINT8},
                {name: "event", parameterType: ParameterType.UINT8},
                {name: "status", parameterType: ParameterType.UINT8},
                {name: "waittime", parameterType: ParameterType.UINT8},
                {name: "suite", parameterType: ParameterType.UINT16},
            ],
        },
        {
            name: "gpioSetDirection",
            ID: 20,
            type: CommandType.SREQ,
            request: [
                {name: "port", parameterType: ParameterType.UINT8},
                {name: "bit", parameterType: ParameterType.UINT8},
                {name: "direction", parameterType: ParameterType.UINT8},
            ],
            response: [
                {name: "oldp0dir", parameterType: ParameterType.UINT8},
                {name: "oldp1dir", parameterType: ParameterType.UINT8},
                {name: "oldp2dir", parameterType: ParameterType.UINT8},
                {name: "p0dir", parameterType: ParameterType.UINT8},
                {name: "p1dir", parameterType: ParameterType.UINT8},
                {name: "p2dir", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "gpioRead",
            ID: 21,
            type: CommandType.SREQ,
            request: [],
            response: [
                {name: "p0", parameterType: ParameterType.UINT8},
                {name: "p1", parameterType: ParameterType.UINT8},
                {name: "p2", parameterType: ParameterType.UINT8},
                {name: "p0dir", parameterType: ParameterType.UINT8},
                {name: "p1dir", parameterType: ParameterType.UINT8},
                {name: "p2dir", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "gpioWrite",
            ID: 22,
            type: CommandType.SREQ,
            request: [
                {name: "port", parameterType: ParameterType.UINT8},
                {name: "bit", parameterType: ParameterType.UINT8},
                {name: "value", parameterType: ParameterType.UINT8},
            ],
            response: [
                {name: "oldp0", parameterType: ParameterType.UINT8},
                {name: "oldp1", parameterType: ParameterType.UINT8},
                {name: "oldp2", parameterType: ParameterType.UINT8},
                {name: "p0", parameterType: ParameterType.UINT8},
                {name: "p1", parameterType: ParameterType.UINT8},
                {name: "p2", parameterType: ParameterType.UINT8},
                {name: "p0dir", parameterType: ParameterType.UINT8},
                {name: "p1dir", parameterType: ParameterType.UINT8},
                {name: "p2dir", parameterType: ParameterType.UINT8},
            ],
        },
        {
            name: "srngGen",
            ID: 76,
            type: CommandType.SREQ,
            request: [],
            response: [{name: "outrng", parameterType: ParameterType.BUFFER100}],
        },
        {
            name: "bindAddEntry",
            ID: 77,
            type: CommandType.SREQ,
            request: [
                {name: "addrmode", parameterType: ParameterType.UINT8},
                {name: "dstaddr", parameterType: ParameterType.IEEEADDR},
                {name: "dstendpoint", parameterType: ParameterType.UINT8},
                {name: "numclusterids", parameterType: ParameterType.UINT8},
                {name: "clusterids", parameterType: ParameterType.LIST_UINT16},
            ],
            response: [
                {name: "srcep", parameterType: ParameterType.UINT8},
                {name: "dstgroupmode", parameterType: ParameterType.UINT8},
                {name: "dstidx", parameterType: ParameterType.UINT16},
                {name: "dstep", parameterType: ParameterType.UINT8},
                {name: "numclusterids", parameterType: ParameterType.UINT8},
                {name: "clusterids", parameterType: ParameterType.BUFFER8},
            ],
        },
    ],
    Subsystem.DEBUG: [
        {
            name: "setThreshold",
            ID: 0,
            type: CommandType.SREQ,
            request: [
                {name: "componentid", parameterType: ParameterType.UINT8},
                {name: "threshold", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "msg",
            ID: 128,
            type: CommandType.AREQ,
            request: [
                {name: "length", parameterType: ParameterType.UINT8},
                {name: "string", parameterType: ParameterType.BUFFER},
            ],
        },
    ],
    Subsystem.APP: [
        {
            name: "msg",
            ID: 0,
            type: CommandType.SREQ,
            request: [
                {name: "appendpoint", parameterType: ParameterType.UINT8},
                {name: "destaddress", parameterType: ParameterType.UINT16},
                {name: "destendpoint", parameterType: ParameterType.UINT8},
                {name: "clusterid", parameterType: ParameterType.UINT16},
                {name: "msglen", parameterType: ParameterType.UINT8},
                {name: "message", parameterType: ParameterType.BUFFER},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "userTest",
            ID: 1,
            type: CommandType.SREQ,
            request: [
                {name: "srcep", parameterType: ParameterType.UINT8},
                {name: "commandid", parameterType: ParameterType.UINT16},
                {name: "param1", parameterType: ParameterType.UINT16},
                {name: "param2", parameterType: ParameterType.UINT16},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "zllTlInd",
            ID: 129,
            type: CommandType.AREQ,
            request: [
                {name: "nwkaddr", parameterType: ParameterType.UINT16},
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "profileid", parameterType: ParameterType.UINT16},
                {name: "deviceid", parameterType: ParameterType.UINT16},
                {name: "version", parameterType: ParameterType.UINT8},
            ],
        },
    ],
    Subsystem.APP_CNF: [
        {
            name: "bdbStartCommissioning",
            ID: 5,
            type: CommandType.SREQ,
            request: [{name: "mode", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "bdbSetChannel",
            ID: 8,
            type: CommandType.SREQ,
            request: [
                {name: "isPrimary", parameterType: ParameterType.UINT8},
                {name: "channel", parameterType: ParameterType.UINT32},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
        {
            name: "bdbComissioningNotifcation",
            ID: 128,
            type: CommandType.AREQ,
            request: [{name: "status", parameterType: ParameterType.UINT8}],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
    ],
    Subsystem.GREENPOWER: [
        {
            name: "secReq",
            ID: 3,
            type: CommandType.SREQ,
            request: [
                {name: "applicationID", parameterType: ParameterType.UINT8},
                {name: "srcID", parameterType: ParameterType.UINT32},
                {name: "gdpIeeeAddr", parameterType: ParameterType.IEEEADDR},
                {name: "endpoint", parameterType: ParameterType.UINT8},
                {name: "gpdfSecurityLevel", parameterType: ParameterType.UINT8},
                {name: "gpdfSecurityFrameCounter", parameterType: ParameterType.UINT8},
                {name: "dgpStubHandle", parameterType: ParameterType.UINT8},
            ],
            response: [{name: "status", parameterType: ParameterType.UINT8}],
        },
    ],
}



from collections.abc import Iterable

import zigpy.types
from zigpy_cc.exception import TODO
from zigpy_cc.types import AddressMode, ParameterType


class BuffaloOptions:
    def __init__(self) -> None:
        self.startIndex = None
        self.length = None


class Buffalo:
    def __init__(self, buffer, position=0) -> None:
        self.position = position
        self.buffer = buffer
        self._len = len(buffer)

    def __len__(self) -> int:
        return len(self.buffer)

    def write_parameter(self, type, value, options):
        if type == ParameterType.UINT8:
            self.write(value)
        elif type == ParameterType.UINT16:
            self.write(value, 2)
        elif type == ParameterType.UINT32:
            self.write(value, 4)
        elif type == ParameterType.IEEEADDR:
            if isinstance(value, Iterable):
                for i in value:
                    self.write(i)
            else:
                self.write(value, 8)
        elif type == ParameterType.BUFFER:
            self.buffer += value
        elif type == ParameterType.LIST_UINT8:
            for v in value:
                self.write(v)
        elif type == ParameterType.LIST_UINT16:
            for v in value:
                self.write(v, 2)
        elif type == ParameterType.LIST_NEIGHBOR_LQI:
            for v in value:
                self.write_neighbor_lqi(v)
        else:
            raise TODO(
                "write %s, value: %s, options: %s", ParameterType(type), value, options
            )

    def write(self, value, length=1, signed=False):
        self.buffer += value.to_bytes(length, "little", signed=signed)

    def write_neighbor_lqi(self, value):
        for i in value["extPanId"]:
            self.write(i)
        for i in value["extAddr"]:
            self.write(i)
        self.write(value["nwkAddr"], 2)
        self.write(
            value["deviceType"]
            | (value["rxOnWhenIdle"] * 4)
            | (value["relationship"] * 16)
        )
        self.write(value["permitJoin"])
        self.write(value["depth"])
        self.write(value["lqi"])

    def read_parameter(self, name, type, options):

        if type == ParameterType.UINT8:
            res = self.read_int()
            if name.endswith("addrmode"):
                res = AddressMode(res)
        elif type == ParameterType.UINT16:
            res = self.read_int(2)
            if (
                name.endswith("addr")
                or name.endswith("address")
                or name.endswith("addrofinterest")
            ):
                res = zigpy.types.NWK(res)
        elif type == ParameterType.UINT32:
            res = self.read_int(4)
        elif type == ParameterType.IEEEADDR:
            res = self.read_ieee_addr()
        elif ParameterType.is_buffer(type):
            type_name = ParameterType(type).name
            length = int(type_name.replace("BUFFER", "") or options.length)
            res = self.read(length)
        elif type == ParameterType.INT8:
            res = self.read_int(signed=True)
        else:
            # list types
            res = []
            if type == ParameterType.LIST_UINT8:
                for i in range(0, options.length):
                    res.append(self.read_int())
            elif type == ParameterType.LIST_UINT16:
                for i in range(0, options.length):
                    res.append(self.read_int(2))
            elif type == ParameterType.LIST_NEIGHBOR_LQI:
                for i in range(0, options.length):
                    res.append(self.read_neighbor_lqi())
            else:
                raise TODO("read type %d", type)

        return res

    def read_int(self, length=1, signed=False):
        return int.from_bytes(self.read(length), "little", signed=signed)

    def read(self, length=1):
        if self.position + length > self._len:
            raise OverflowError
        res = self.buffer[self.position : self.position + length]
        self.position += length
        return res

    def read_ieee_addr(self):
        return zigpy.types.EUI64(self.read(8))

    def read_neighbor_lqi(self):
        item = dict()
        item["extPanId"] = self.read_ieee_addr()
        item["extAddr"] = self.read_ieee_addr()
        item["nwkAddr"] = zigpy.types.NWK(self.read_int(2))

        value1 = self.read_int()
        item["deviceType"] = value1 & 0x03
        item["rxOnWhenIdle"] = (value1 & 0x0C) >> 2
        item["relationship"] = (value1 & 0x70) >> 4

        item["permitJoin"] = self.read_int() & 0x03
        item["depth"] = self.read_int()
        item["lqi"] = self.read_int()

        return item
